#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Thermodynamics utilities
"""
import numpy as np

import xarray as xr
from .__init__ import XoaError, xoa_warn
from . import cf as xcf
from . import misc as xmisc
from . import coords as xcoords
from . import regrid as xregrid


def _get_array_(ds, func, variants, variant, errors, name):
    if not variant:
        variant = None
    if isinstance(variant, str) or variant is None:
        variant = [variant]

    das = []
    for da in ds.values():
        for vr in variant:
            if func(da, variant):
                das.append(da)

    errors = xmisc.ERRORS[errors]
    if not das:
        msg = f"Found no {name} array"
        if errors == "raise":
            raise XoaError(msg)
        if errors == "warning":
            xoa_warn(msg)
        return
    if len(das) > 1:
        msg = f"Found more than one {name} array"
        if errors == "raise":
            raise XoaError(msg)
        if errors == "warning":
            xoa_warn(msg, stacklevel=3)
    return das[0]


TEMP_VARIANTS = xmisc.Choices(
    {
        None: "No restriction",
        "insitu": "In situ",
        "conservative": "Conservative temperature",
        "absolute": "Absolute temperature",
    },
    parameter="variant",
    description="Restrict checking to a given variant",
)


@TEMP_VARIANTS.format_method_docstring
def is_temp(da, variant=None):
    """Check if `da` is a temperature-like array

    Parameters
    ----------
    da: xarray.DataArray
    {variant}

    Return
    ------
    bool
    """
    cfspecs = xcf.get_cf_specs(da)
    variant = TEMP_VARIANTS[variant]
    cf_names = []
    if variant is None or variant == "insitu":
        cf_names.append("temp")
    if variant is None or variant == "potential":
        cf_names.append("ptemp")
    if variant is None or variant == "conservative":
        cf_names.append("ctemp")
    for cf_name in cf_names:
        if cfspecs.match_data_var(da, cf_name):
            return True
    return False


@xmisc.ERRORS.format_function_docstring
def get_temp(ds, variant=None, errors="warn"):
    """Search for temperature in a dataset

    Parameters
    ----------
    da: xarray.Dataset
    variant: None, str, list(str)
        Variant of temperature or list of them. See :func:`is_temp`.
    {errors}

    Return
    ------
    xarray.DataArray, None
        If None or several arrays are found, a warning or an error may be raised
        depending on the `errors` parameter.
        If several arrays are matching and `errors` is "warning", the first array is returned.

    """
    return _get_array_(ds, is_temp, TEMP_VARIANTS, variant, errors, "temperature")


SAL_VARIANTS = xmisc.Choices(
    {
        None: "No restriction",
        "insitu": "In situ salinity",
        "absolute": "Absolute salinity",
        "preformed": "Preformed salinity",
        "practical": "Practical salinity",
    },
    parameter="variant",
    description="Restrict checking to a given variant",
)


@SAL_VARIANTS.format_method_docstring
def is_sal(da, variant=None):
    """Check if `da` is a salinity-like array.

    Parameters
    ----------
    da: xarray.DataArray
    {variant}

    Return
    ------
    bool
    """
    cfspecs = xcf.get_cf_specs(da)
    cf_names = []
    if variant is None or variant == "insitu":
        cf_names.append("sal")
    if variant is None or variant == "practical":
        cf_names.append("psal")
    if variant is None or variant == "preformed":
        cf_names.append("pfsal")
    if variant is None or variant == "absolute":
        cf_names.append("asal")
    for cf_name in cf_names:
        if cfspecs.match_data_var(da, cf_name):
            return True
    return False


@xmisc.ERRORS.format_function_docstring
def get_sal(ds, variant=None, errors="warn"):
    """Search for salinity in a dataset.

    Parameters
    ----------
    da: xarray.Dataset
    variant: None, str, list(str)
        Variant of salinity or list of them. See :func:`is_sal`.
    {errors}

    Return
    ------
    xarray.DataArray, None
        Return None if not found
    """
    return _get_array_(ds, is_sal, SAL_VARIANTS, variant, errors, "salinity")


DENS_VARIANTS = xmisc.Choices(
    {
        None: "No restriction",
        "insitu": "In situ density",
        "potential": "Potential density",
        "neutral": "Neutral salinity",
    },
    parameter="variant",
    description="Restrict checking to a given variant",
)


@DENS_VARIANTS.format_method_docstring
def is_dens(da, variant=None):
    """Check if `da` is a density-like array.

    Parameters
    ----------
    da: xarray.DataArray
    {variant}

    Return
    ------
    bool
    """
    cfspecs = xcf.get_cf_specs(da)
    cf_names = []
    if variant is None or variant == "insitu":
        cf_names.extend(["dens", "sigmat"])
    if variant is None or variant == "potential":
        cf_names.extend(["pdens", "sigmatheta", "sigma0", "sigma1", "sigma2", "sigma3", "sigma4"])
    if variant is None or variant == "neutral":
        cf_names.append("ndens")
    for cf_name in cf_names:
        if cfspecs.match_data_var(da, cf_name):
            return True
    return False


@xmisc.ERRORS.format_function_docstring
def get_dens(ds, variant=None, errors="warn"):
    """Search for density in a dataset.

    Parameters
    ----------
    da: xarray.Dataset
    variant: None, str, list(str)
        Variant of density or list of them. See :func:`is_dens`.
    {errors}

    Return
    ------
    xarray.DataArray, None
        Return None if not found
    """
    return _get_array_(ds, is_dens, DENS_VARIANTS, variant, errors, "density")


MLD_METHODS = xmisc.Choices(
    {
        "deltatemp": (
            "depth at which the potential temperature is `deltadtemp` "
            "lower than the surface temperature"
        ),
        "deltadens": (
            "depth at which the potential density is `deltadens` higher than the surface density"
        ),
        "kzmax": "depth at which the vertical diffusivity value reaches the `kzmax` value",
    },
    parameter="method",
    description="Method for computing the mixed layer depth",
)


@MLD_METHODS.format_method_docstring
def mixed_layer_depth(
    da,
    method=None,
    zref=0.0,
    deltatemp=0.2,
    deltadens=0.3,
    kzmax=0.0005,
    zdim=None,
    dask="parallelized",
    **kwargs,
):
    """Compute the mixed layer depth with different methods.

    Parameters
    ----------
    da: xarray.DataArray
        A data array that contains either the potential temperature, the potential density or
        the vertical tracer diffisivity.
        Note that this array **must contain a depth coordinate**, which optimally contains
        the `positive` attribute.
    {method}
    zref: float
        Reference depth (in meters) from which the MLD_METHODS are applied
    kwargs: dict
        Extra keywords are passed to :func:`xoa.regrid.isoslice`.

    Raise
    -----
    XoaError:
        When `method` is `None` and cannot be inferred from `da`.

    Return
    ------
    xarray.DataArray
        Mixed layer depth

    See Also
    --------
    xoa.regrid.isoslice
    """
    # Vertical dimension
    if zdim is None:
        zdim = xcoords.get_zdim(da)
    else:
        assert zdim in da.dims
    assert zdim in da.dims

    # Depths
    depth = xcoords.get_depth(da)
    positive = xcoords.get_positive_attr(depth, zdim=zdim) or "up"

    # Method
    cfspecs = xcf.get_cf_specs(da)
    if method is None:
        if is_temp(da):
            method = "deltatemp"
        elif cfspecs.match_data_var(da, 'kz'):
            method = "kzmax"
        elif is_dens(da):
            method = "deltadens"
        else:
            raise XoaError(
                "Cannot infer mixed layer depth computation method from data array. "
                "Please specify the `method` keyword."
            )
    else:
        method = MLD_METHODS[method]

    # Iso value
    if method == "kzmax":
        isoval = kzmax
    else:
        if zref == 0.0:
            surf = da.isel({zdim: 0 if positive == "down" else -1})
        else:
            dep0 = xr.DataArray([zref], dims="depth")
            surf = xregrid.regrid1d(da, dep0, method="linear").squeeze(dim="depth")
        if method == "deltatemp":
            isoval = surf - deltatemp
        elif method == "deltadens":
            isoval = surf + deltadens

    # Slice
    mld = xregrid.isoslice(depth, da, isoval, dim=zdim, dask=dask, **kwargs)
    mld = np.abs(mld)
    mld = mld.drop_vars(depth.name, errors="ignore")

    # Format
    mld.attrs = {}
    cfspecs.format_data_var(
        mld, "mld", format_coords=False, rename_dims=False, copy=False, replace_attrs=True
    )

    return mld


# def get_mld(ds, methods=[None, "deltadens", "deltatemp", "kz"], **kwargs):
#     """Get or compute the mixed layer depth from a dataset"""
#     cfspecs = xcf.get_cf_specs(ds)
#     for method in methods:
#         if method is None:
#             mld = cfspecs.search_data_var(ds, "mld", errors="ignore")
#             if mld is None:



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.sigma` module
"""
import numpy as np
import xarray as xr

import xoa
from xoa import sigma


def test_sigma_get_cs():

    nz = 3
    sig = xr.DataArray(np.linspace(-1, 0, nz), dims="sig")
    thetas = 7.0
    thetab = 2.0
    cs = sigma.get_cs(sig, thetas, thetab, cs_type=None)
    assert cs.dims == sig.dims
    np.testing.assert_allclose(cs, [-1.0, -0.96983013, 0.0])


def test_sigma_atmosphere_sigma_coordinate():
    nt, nz, nx = 2, 3, 5
    sig = xr.DataArray(np.linspace(-1, 0, nz), dims="sig")
    ptop = xr.DataArray(np.ones((nt, nx)), dims=("nt", "nx"))
    ps = xr.DataArray(150 * np.ones(nx), dims="nx")
    sigma.atmosphere_sigma_coordinate(sig, ptop, ps)


def test_ocean_sigma_coordinate():
    nt, nz, nx = 2, 3, 5
    sig = xr.DataArray(np.linspace(-1, 0, nz), dims="sig")
    ssh = xr.DataArray(np.ones((nt, nx)), dims=("nt", "nx"))
    bathy = xr.DataArray(150 * np.ones(nx), dims="nx")
    sigma.ocean_sigma_coordinate(sig, ssh, bathy)


def test_ocean_s_coordinate():
    nt, nz, nx = 2, 3, 5
    sig = xr.DataArray(np.linspace(-1, 0, nz), dims="sig")
    ssh = xr.DataArray(np.ones((nt, nx)), dims=("nt", "nx"))
    bathy = xr.DataArray(150 * np.ones(nx), dims="nx")
    hc = 10.0 + bathy * 0
    thetas = 7.0
    thetab = 2.0
    sigma.ocean_s_coordinate(sig, ssh, bathy, hc, thetas, thetab)


def test_ocean_s_coordinate_g1():
    nt, nz, nx = 2, 3, 5
    sig = xr.DataArray(np.linspace(-1, 0, nz), dims="sig")
    ssh = xr.DataArray(np.ones((nt, nx)), dims=("nt", "nx"))
    bathy = xr.DataArray(150 * np.ones(nx), dims="nx")
    hc = 10.0 + bathy * 0
    thetas = 7.0
    thetab = 2.0
    sigma.ocean_s_coordinate_g1(sig, ssh, bathy, hc, thetas, thetab)


def test_ocean_s_coordinate_g2():
    nt, nz, nx = 2, 3, 5
    sig = xr.DataArray(np.linspace(-1, 0, nz), dims="sig")
    ssh = xr.DataArray(np.ones((nt, nx)), dims=("nt", "nx"))
    bathy = xr.DataArray(150 * np.ones(nx), dims="nx")
    hc = 10.0 + bathy * 0
    thetas = 7.0
    thetab = 2.0
    sigma.ocean_s_coordinate_g2(sig, ssh, bathy, hc, thetas, thetab)


def test_sigma_decode_cf_sigma():

    ds = xoa.open_data_sample("croco.south-africa.meridional.nc")
    dsd = sigma.decode_cf_sigma(ds)

    assert "depth" in dsd
    assert dsd.depth.dims == ('time', 's_rho', 'eta_rho', 'xi_rho')
    assert dsd.depth.shape == (1, 32, 56, 1)



"""
Low level numeric utilities

The numerical inputs and outputs of all these routines are of scalar
or numpy.ndarray type.
"""
# Copyright 2020-2022 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import numpy as np
import numba

NOT_CI = os.environ.get("CI", "false") == "false"


@numba.njit(cache=NOT_CI)
def get_iminmax(data1d):
    """The first and last non nan values for a 1d array

    Parameters
    ----------
    data1d: array_like(n)

    Return
    ------
    int
        Index of the first valid value
    int
        Index of the last valid value
    """
    imin = -1
    imax = -1
    n = len(data1d)
    for i in range(n):
        if imin < 0 and not np.isnan(data1d[i]):
            imin = i
        if imax < 0 and not np.isnan(data1d[n - 1 - i]):
            imax = n - 1 - i
        if imax > 0 and imin > 0:
            break
    return imin, imax


@numba.njit(numba.int64[:](numba.int64, numba.int64[:]), cache=NOT_CI)
def unravel_index(i, shape):

    ir = i
    ndim = len(shape)
    ii = np.zeros(ndim, np.int64)
    for o in range(ndim):
        if o != ndim - 1:
            base = np.prod(shape[o + 1 :])
        else:
            base = 1
        ii[o] = ir // base
        ir -= ii[o] * base
        # print(o, base, ir)
    return ii


@numba.njit(numba.int64(numba.int64[:], numba.int64[:]), cache=NOT_CI)
def ravel_index(ii, shape):
    ir = 0
    ndim = len(shape)
    for o in range(ndim):
        if o != ndim - 1:
            base = np.prod(shape[o + 1 :])
        else:
            base = 1
        # print(ii[o], base)
        ir += ii[o] * base
    return ir



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Geographic utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import numpy as np
import numba
import xarray as xr

from . import misc
from . import coords as xcoords


#: Earth radius in meters
EARTH_RADIUS = 6371e3


@numba.vectorize(cache=True)
def _haversine_(lon0, lat0, lon1, lat1):
    """Haversine distance between two points on a **unit sphere**

    Parameters
    ----------
    lon0: float, array_like
        Longitude of the first point(s)
    lat0: float, array_like
        Latitude of the first point(s)
    lon1: float, array_like
        Longitude of the second point(s)
    lat1: float, array_like
        Latitude of the second point(s)

    Return
    ------
    float, array_like
        Distance(s)
    """
    deg2rad = math.pi / 180.0
    dist = math.sin(deg2rad * (lat0 - lat1) * 0.5) ** 2
    dist += (
        math.cos(deg2rad * lat0)
        * math.cos(deg2rad * lat1)
        * math.sin(deg2rad * (lon0 - lon1) * 0.5) ** 2
    )
    dist = 2.0 * math.asin(math.sqrt(dist))
    return dist


def haversine(lon0, lat0, lon1, lat1, radius=EARTH_RADIUS):
    """Haversine distance between two points

    Parameters
    ----------
    lon0: float, array_like
        Longitude of the first point(s)
    lat0: float, array_like
        Latitude of the first point(s)
    lon1: float, array_like
        Longitude of the second point(s)
    lat1: float, array_like
        Latitude of the second point(s)
    radius: float
        Radius of the sphere which defaults to the earth radius

    Return
    ------
    float, array_like
        Distance(s)
    """
    return _haversine_(np.double(lon0), np.double(lat0), np.double(lon1), np.double(lat1)) * radius


def bearing(lon0, lat0, lon1, lat1):
    """Compute the bearing angle (forward azimuth)

    Parameters
    ----------
    lon0: float, array_like
        Longitude of the first point(s)
    lat0: float, array_like
        Latitude of the first point(s)
    lon1: float, array_like
        Longitude of the second point(s)
    lat1: float, array_like
        Latitude of the second point(s)

    Return
    ------
    float, array_like
        Angle(s)
    """
    return _bearing_(np.double(lon0), np.double(lat0), np.double(lon1), np.double(lat1))


@numba.vectorize
def _bearing_(lon0, lat0, lon1, lat1):
    deg2rad = math.pi / 180.0
    a = math.arctan2(
        math.cos(deg2rad * lat0) * math.sin(deg2rad * lat1)
        - math.sin(deg2rad * lat0) * math.cos(deg2rad * lat1) * math.cos(deg2rad * (lon1 - lon0)),
        math.sin(deg2rad * (lon1 - lon0)) * math.cos(deg2rad * lat1),
    )
    return a * 180 / math.pi


class distance_units(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Supported units of distance"""

    #: Meters (default)
    m = 0
    #: Meters (default)
    meters = 0
    #: Meters (default)
    meter = 0
    #: Kilometers
    km = 1
    #: Kilometers
    kilometers = 1
    #: Kilometers
    kilometer = 1


def get_distances(da0, da1=None, radius=EARTH_RADIUS, units="m"):
    """Compute the haversine distance between two datasets/data arrays

    Parameters
    ----------
    da0: xarray.Dataset, xarray.DataArray
    da1: xarray.Dataset, xarray.DataArray
    radius: float
        Radius of the sphere which defaults to the earth radius
    units: int, str, distance_units
        Distance units as one of: {distance_units.rst_with_links}

    Return
    ------
    xarray.DataArray
        An array with dims `(npts0, npts1)`
    See also
    --------
    haversine
    cdist
    pdist
    """
    ds0 = xcoords.geo_stack(da0, rename=True, drop=True, reset_index=True)
    XY0 = np.dstack([ds0.lon.values, ds0.lat.values])[0]
    units = distance_units[units]
    du = str(units)
    import xarray as xr

    if da1 is None:
        dd = pdist(XY0, radius=radius)
        if units == distance_units.km:
            dd *= 1e-3
        dd = xr.DataArray(
            dd,
            dims=("npts", "npts"),
            attrs={"units": du},
            name="distance",
            coords={"lon": ds0.lon, "lat": ds0.lat},
        )
    else:
        ds0 = ds0.rename(npts="npts0")
        ds1 = xcoords.geo_stack(da1, "npts1", rename=True, drop=True, reset_index=True)
        XY1 = np.dstack([ds1.lon.values, ds1.lat.values])[0]
        dd = cdist(XY0, XY1, radius=radius)
        if units == distance_units.km:
            dd *= 1e-3
        dd = xr.DataArray(
            dd,
            dims=("npts0", "npts1"),
            attrs={"units": du},
            name="distance",
            coords={"lon_0": ds0.lon, "lat_0": ds0.lat, "lon_1": ds1.lon, "lat_1": ds1.lat},
        )
    return dd


get_distances.__doc__ = get_distances.__doc__.format(**locals())


def cdist(XA, XB, radius=EARTH_RADIUS, **kwargs):
    """A scipy-distance like cdist function for the haversine method

    Parameters
    ----------
    XA: numpy.array
        An ma by 2 array of coordinates of the first dataset in a geographical space.
    XB: numpy.array
        An mb by 2 array of coordinates of the second dataset in a geographical space.
    radius: float
        Radius of the sphere which defaults to the earth radius

    Returns
    -------
    numpy.array
        2D array of distances

    See also
    --------
    haversine
    pdist
    scipy.sparial.distances.cdist
    scipy.sparial.distances.pdist
    """
    lons0 = XA[:, 0]
    lats0 = XA[:, 1]
    lons1 = XB[:, 0]
    lats1 = XB[:, 1]
    xx = np.meshgrid(lons1, lons0)
    yy = np.meshgrid(lats1, lats0)
    return haversine(xx[0], yy[0], xx[1], yy[1], radius=radius)


def pdist(X, compact=False, radius=EARTH_RADIUS, **kwargs):
    """A scipy-distance like pdist function for the haversine method

    Parameters
    ----------
    X: numpy.array
        An m by 2 array of coordinates in a geographical space.
    compact: bool
        Compact the distance matrix to remove duplicate and zeros.
        It is the strict upper triangle of the distance matrix.
    radius: float
        Radius of the sphere which defaults to the earth radius

    Returns
    -------
    numpy.array
        Either 2D (square form) or 1D (compact form) the distance matrix

    See also
    --------
    haversine
    cdist
    numpy.triu
    scipy.sparial.distances.pdist
    scipy.sparial.distances.cdist
    scipy.sparial.distances.squareform
    """
    dd = cdist(X, X, radius=radius)
    if compact:
        return dd[np.triu_indices(dd.shape[0], 1)]
    return dd


def _adapative_cdist_(XA, XB, method="haversine", **kwargs):
    if method == "haversine":
        return cdist(XA, XB, **kwargs)
    import scipy.spatial.distance

    return scipy.spatial.distance.cdist(XA, XB, method=method, **kwargs)


def _adapative_pdist_(X, method="haversine", **kwargs):
    if method == "haversine":
        return pdist(X, **kwargs)
    import scipy.spatial.distance

    return scipy.spatial.distance.pdist(X, method=method, **kwargs)


class ScipyDistContext(object):
    """Context to switch the :func:`scipy.spatial.distance.cdist` fonction to :func:`cdist`

    Parameters
    ----------
    force: bool
        If true, this function will be used whatever the distance method asked for is.
    """

    def __init__(self, cdist=True, pdist=True, force=False):
        self.switch_cdist = cdist
        self.switch_pdist = pdist
        import scipy.spatial.distance

        self.distmod = scipy.spatial.distance
        if force:
            self.cdist = cdist
            self.pdist = pdist
        else:
            self.cdist = _adapative_cdist_
            self.pdist = _adapative_pdist_

    def __enter__(self):
        if self.switch_cdist:
            self._old_cdist = getattr(self.distmod, "cdist")
            setattr(self.distmod, "cdist", self.cdist)
        if self.switch_pdist:
            self._old_pdist = getattr(self.distmod, "pdist")
            setattr(self.distmod, "pdist", self.pdist)
        return self

    def __exit__(self, type, value, traceback):
        if self.switch_cdist:
            setattr(self.distmod, "cdist", self._old_cdist)
        if self.switch_pdist:
            setattr(self.distmod, "pdist", self._old_pdist)


def get_extent(extent, margin=0, square=False, min_extent=None):
    """Compute the geographic extent in degrees

    Parameters
    ----------
    extent: xarray.DataArray, xarray.Dataset, dict, tuple, list
        Either:

        - An array or dataset with longitude and latitude coordinates.
        - A dict with ``lon`` and ``lat`` keys: ``dict(lon=..., lat=...)``
        - A two-element tuple of longitudes and latitudes: ``(lon, lat)``
        - A extent list: ``[xmin, xmax, ymin, ymax]``.
    margin: float
        A relative fraction of the width and height that is used to set margins.
        For instance, a value of ``-0.1`` shrinks the box of 10% on each side.
    square: bool
        Force the box to be square in degrees.
    min_extent: None, float, (float, float)
        Minimal extent along x and y: ``(dx, dy)``.
        If a single floating value is provided, it is valid for both x and y.

    Return
    ------
    list
        ``[xmin, xmax, ymin, ymax]``

    Example
    -------

    .. ipython:: python

        @suppress
        from xoa.geo import get_extent
        @suppress
        import numpy as np
        get_extent([10., 20., 10., 20.], margin=0.1)
        get_extent({"lon": np.linspace(10, 20, 5), "lat": np.linspace(10, 20, 5)}, square=True)
        get_extent((np.linspace(10, 20, 5), np.linspace(10, 20, 5)), margin=-.1, square=True)

    """
    # Get min and max
    if hasattr(extent, "coords"):
        lon = xcoords.get_lon(extent)
        lat = xcoords.get_lat(extent)
        extent = (lon, lat)
    elif isinstance(extent, dict):
        extent = extent['lon'], extent['lat']
    if isinstance(extent, (list, np.ndarray)):
        xmin, xmax, ymin, ymax = extent
    else:  # tuple
        lon, lat = extent
        xmin = float(np.min(lon))
        xmax = float(np.max(lon))
        ymin = float(np.min(lat))
        ymax = float(np.max(lat))

    # Scale
    if square or margin or min_extent is not None:
        dx = xmax - xmin
        dy = ymax - ymin
        x0 = 0.5 * (xmin + xmax)
        y0 = 0.5 * (ymin + ymax)
        if min_extent is not None:
            if isinstance(min_extent, (float, int)):
                min_extent = (min_extent, min_extent)
            dx_min, dy_min = min_extent
            dx = max(dx, dx_min)
            dy = max(dy, dy_min)
        if square and dy != 0:
            aspect = dx * math.cos(y0 * math.pi / 180) / dy
            if aspect > 1:
                dy *= aspect
            else:
                dx /= aspect
        xmargin = margin * dx
        ymargin = margin * dy
        xmin = x0 - 0.5 * dx - xmargin
        xmax = x0 + 0.5 * dx + xmargin
        ymin = y0 - 0.5 * dy - ymargin
        ymax = y0 + 0.5 * dy + ymargin

    return [xmin, xmax, ymin, ymax]


def clusterize(obj, npmax, split=False):
    """Split data into clouds of points of max size `npmax`

    Input object must have valid geographic coordinates.

    Parameters
    ----------
    obj: xarray.DataArray, xarray.Dataset
        If a data array, it must have valid longitude and latitude coordinates.
    npmax: int
        Maximal number of point per cluster
    split:
        Return one dataset per cluster

    Return
    ------
    xarray.Dataset, list of xarray.Dataset
        A dataset has its longitude and latitude coordinates renamed `lon` and `lat`,
        and its stacked dimension renamed `npts`.
        It contains only arrays that contains the `npts` dimension.
        If a clustering was needed, the dataset contains the following arrays:

        iclust:
            Index of the cluster points belongs to.
        indices:
            Indices in the original dataset.
        centroid:
            Coordinate of the centroid(s)
        distorsion:
            Distances to the centroid that the points belongs to.

        If the input is a dataset, the global attribute :attr:`var_names`
        is set to the list of data var names.


    """
    # Positions
    obj = xcoords.geo_stack(obj, "npts", rename=True, drop=True)
    x = obj.lon.values
    y = obj.lat.values

    from scipy.cluster.vq import kmeans, vq

    # Nothing to do
    csize = len(x)
    if not isinstance(obj, xr.Dataset):
        obj = obj.to_dataset(name=obj.name or "data")
    obj.encoding["clust_var_names"] = list(obj)
    if npmax < 2 or csize <= npmax:
        return [obj] if split else obj

    # Loop on the number of clusters
    nclust = 2
    points = np.dstack((x, y))[0]
    ii = np.arange(csize)
    while csize > npmax:
        centroids, global_distorsion = kmeans(points, nclust)
        indices, distorsions = vq(points, centroids)
        sindices = [ii[indices == nc] for nc in range(nclust)]
        csizes = [sii.shape[0] for sii in sindices]
        order = np.argsort(csizes)[::-1]
        csize = csizes[order[0]]
        sdistorsions = [distorsions[sii] for sii in sindices]
        nclust += 1
    nclust = len(centroids)

    if split:

        #  Reorder
        sindices = [sindices[i] for i in order]
        sdistorsions = [sdistorsions[i] for i in order]
        centroids = centroids[order]

        # Split
        out = []
        for ic in range(nclust):
            obji = obj.isel(npts=sindices[ic])
            obji["iclust"] = xr.DataArray(ic)
            obji["indices"] = xr.DataArray(sindices[ic], dims="npts")
            obji["centroid"] = xr.DataArray(centroids[ic], dims="xy")
            obji["distorsion"] = xr.DataArray(sdistorsions[ic], dims="npts")
            obji.attrs["global_distorsion"] = global_distorsion
            obji.attrs["npmax"] = npmax
            out.append(obji)
        return out

    obj["iclust"] = xr.DataArray(indices, dims="npts")
    obj["indices"] = xr.DataArray(ii, dims="npts")
    obj["centroid"] = xr.DataArray(centroids, dims=("nclust", "xy"))
    obj["distorsion"] = xr.DataArray(distorsions, dims="npts")
    obj.attrs["global_distorsion"] = global_distorsion
    obj.attrs["npmax"] = npmax
    return obj


def deg2m(deg, lat=None, radius=EARTH_RADIUS):
    """Convert to meters a zonal or meridional distance in degrees

    Parameters
    ----------
    deg: float
        Longitude step
    lat: float
        Latitude for a zonal distance

    Return
    ------
    float
    """
    dd = deg * np.pi * radius / 180.0
    if lat:
        dd *= np.cos(np.radians(lat))
    return dd


def m2deg(met, lat=None, radius=EARTH_RADIUS):
    """Convert to degrees a zonal or meridional distance in meters

    Parameters
    ----------
    met: float
        Longitude step
    lat: float
        Latitude for a zonal distance

    Return
    ------
    float
    """
    dd = met * 180 / (np.pi * radius)
    if lat:
        dd /= np.cos(np.radians(lat))
    return dd



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.geo` module
"""

import numpy as np
import xarray as xr

from xoa import geo


def test_geo_haversine():

    assert geo.haversine(0.0, 0.0, 180.0, 0.0, 1.0) == np.pi
    assert geo.haversine(0, -90.0, 0, 90.0, 1.0) == np.pi
    assert geo.haversine(0, -90.0, 0, 90.0, 2.0) == 2 * np.pi
    assert int(geo.haversine(0, 0, 10.0, 10.0)) == 1568520


def test_geo_cdist():

    xy0 = np.array([[0, 1], [0.0, 1.0]]).T
    xy1 = np.array([[0, 0.5, 180.0], [0.0, 0.5, 0]]).T
    dd = geo.cdist(xy0, xy1, radius=1)
    assert dd.shape == (2, 3)
    assert dd[0, 0] == 0.0
    assert dd[0, -1] == np.pi


def test_geo_pdist():

    xy0 = np.array([[0, 1], [0.0, 1.0]]).T
    dd = geo.pdist(xy0)
    assert dd.shape == (2, 2)
    ddc = geo.pdist(xy0, compact=True)
    assert ddc.shape == (1,)
    assert ddc[0] == dd[1, 0]


def test_geo_clusterize():

    x = [0, 1, 3, 4, 8, 9, 10.0]
    y = x
    ds = xr.Dataset(coords={"lon": ("npts", x), "lat": ("npts", y)})
    clusters = geo.clusterize(ds, npmax=3, split=True)
    assert len(clusters) == 3
    assert set([c.sizes["npts"] for c in clusters]) == {2, 2, 3}



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.interp` module
"""

import functools
import numpy as np
import pytest

from xoa import interp


def vfunc(t=0, z=0, y=0, x=0):
    """A function that returns a linear combination of coordinates"""
    return 1.13 * x + 2.35 * y + 3.24 * z - 0.65 * t


def round_as_time(arr, units="us", origin="1950-01-01"):
    arr = arr.astype(f"m8[{units}]")
    origin = np.datetime64(origin, units)
    arr = arr + origin
    return (arr - origin) / np.timedelta64(1, units)


@functools.lru_cache()
def get_interp1d_data(
    yimin=-100.0,
    yimax=0.0,
    yomin=-90.0,
    yomax=10.0,
    irregular=True,
    nx=17,
    nyi=15,
    nyo=25,
    mask=True,
):

    np.random.seed(0)

    # coords
    yi = np.linspace(yimin, yimax, nyi)
    yo = np.linspace(yomin, yomax, nyo)
    x = np.linspace(0, 700, nx)
    yyi = np.resize(yi, (nx, nyi))
    yyo = np.resize(yo, (nx, nyo))
    if irregular:
        dyi = (yi[1] - yi[0]) * 0.49
        yyi += np.random.uniform(-dyi, dyi, (nx, nyi))
        dyo = (yo[1] - yo[0]) * 0.49
        yyo += +np.random.uniform(-dyo, dyo, (nx, nyo))
    xxi = np.resize(x, (nyi, nx)).T
    xxo = np.resize(x, (nyo, nx)).T

    # input
    xxi = np.resize(x, (nyi, nx)).T
    vari = vfunc(y=yyi, x=xxi)
    if mask:
        vari[int(nx / 3) : int(2 * nx / 3), int(nyi / 3) : int(2 * nyi / 3)] = np.nan

    # shapes of extra dims
    eshapes = np.vstack((vari.shape[:-1], yyi.shape[:-1], yyo.shape[:-1]))

    return xxi, yyi, vari, xxo, yyo, eshapes


def test_interp_nearest1d():

    # Get data
    xxi, yyi, vari, xxo, yyo, eshapes = get_interp1d_data()

    # Interpolation
    varon = interp.nearest1d(vari, yyi, yyo, eshapes)
    yyon = interp.nearest1d(yyi, yyi, yyo, eshapes)
    xxon = interp.nearest1d(xxi, yyi, yyo, eshapes)
    varon_true = vfunc(y=yyon, x=xxon)
    varon_true[np.isnan(varon)] = np.nan
    np.testing.assert_allclose(varon_true, varon)


def test_interp_linear1d():

    # Get data
    xxi, yyi, vari, xxo, yyo, eshapes = get_interp1d_data()

    # Interpolation
    varol = interp.linear1d(vari, yyi, yyo, eshapes)
    assert not np.isnan(varol).all()
    varol_true = vfunc(y=yyo, x=xxo)
    varol_true[np.isnan(varol)] = np.nan
    np.testing.assert_allclose(varol_true, varol)


def test_interp_cubic1d():

    # Get data
    xxi, yyi, vari, xxo, yyo, eshapes = get_interp1d_data()

    # Interpolation
    varoh = interp.cubic1d(vari, yyi, yyo, eshapes)
    assert not np.isnan(varoh).all()
    assert np.nanmax(varoh) <= np.nanmax(vari)
    assert np.nanmin(varoh) >= np.nanmin(vari)


def test_interp_hermit1d():

    # Get data
    xxi, yyi, vari, xxo, yyo, eshapes = get_interp1d_data()

    # Interpolation
    varoh = interp.hermit1d(vari, yyi, yyo, eshapes)
    assert not np.isnan(varoh).all()
    assert np.nanmax(varoh) <= np.nanmax(vari)
    assert np.nanmin(varoh) >= np.nanmin(vari)


@pytest.mark.parametrize("method", ["nearest", "linear", "cubic", "hermit"])
def test_interp1d_nans_in_coords(method):

    # Get data
    xxi, yyi, vari, xxo, yyo, eshapes = get_interp1d_data()

    # Add nans to coords
    yyin = yyi.copy()
    yyin[:, :3] = np.nan
    yyin[:, -3:] = np.nan
    yyon = yyo.copy()
    yyon[:, :3] = np.nan
    yyon[:, -3:] = np.nan

    # Interpolations
    func = getattr(interp, method + "1d")
    varol = func(vari[:, 3:-3], yyi[:, 3:-3], yyo[:, 3:-3], eshapes)
    varoln = func(vari, yyin, yyon, eshapes)
    np.testing.assert_allclose(varol, varoln[:, 3:-3])


@pytest.mark.parametrize("method", ["nearest", "linear", "cubic", "hermit"])
def test_interp_interp1d_eshapes(method):

    # Get data and func
    xxi, yyi, vari, xxo, yyo, eshapes = get_interp1d_data(nx=18, mask=False, irregular=False)
    eshapes = np.repeat([[3, 6]], 3, axis=0)
    func = getattr(interp, method + "1d")

    # Reference
    varol_ref = func(vari, yyi, yyo, eshapes).reshape(3, 6, -1)

    # missing dim 0 for vari
    vari0 = vari.reshape(3, 6, -1)[0]
    eshapes0 = eshapes.copy()
    eshapes0[0, 0] = 1
    varol0 = func(vari0, yyi, yyo, eshapes0).reshape(3, 6, -1)
    np.testing.assert_allclose(varol_ref[0], varol0[0])
    np.testing.assert_allclose(varol_ref[0], varol0[1])

    # missing dim 1 for vari
    vari1 = vari.reshape(3, 6, -1)[:, 0]
    eshapes1 = eshapes.copy()
    eshapes1[0, 1] = 1
    varol1 = func(vari1, yyi, yyo, eshapes1).reshape(3, 6, -1)
    np.testing.assert_allclose(varol_ref[:, 0], varol1[:, 0])
    np.testing.assert_allclose(varol_ref[:, 0], varol1[:, 1])

    # missing dim 0 for yyi
    yyi0 = yyi.reshape(3, 6, -1)[0]
    eshapes0 = eshapes.copy()
    eshapes0[1, 0] = 1
    varol0 = func(vari, yyi0, yyo, eshapes0).reshape(3, 6, -1)
    np.testing.assert_allclose(varol_ref, varol0)

    # missing dim 1 for yyi
    yyi1 = yyi.reshape(3, 6, -1)[:, 0]
    eshapes1 = eshapes.copy()
    eshapes1[1, 1] = 1
    varol1 = func(vari, yyi1, yyo, eshapes1).reshape(3, 6, -1)
    np.testing.assert_allclose(varol_ref, varol1)

    # missing dim 0 for yyo
    yyo0 = yyo.reshape(3, 6, -1)[0]
    eshapes0 = eshapes.copy()
    eshapes0[2, 0] = 1
    varol0 = func(vari, yyi, yyo0, eshapes0).reshape(3, 6, -1)
    np.testing.assert_allclose(varol_ref, varol0)

    # missing dim 1 for yyo
    yyo1 = yyo.reshape(3, 6, -1)[:, 0]
    eshapes1 = eshapes.copy()
    eshapes1[2, 1] = 1
    varol1 = func(vari, yyi, yyo1, eshapes1).reshape(3, 6, -1)
    np.testing.assert_allclose(varol_ref, varol1)


def test_interp_cellave1d():

    np.random.seed(0)

    # coords
    nx = 17
    nyi = 20
    nyo = 12
    yib = np.linspace(-1000.0, 0.0, nyi + 1)
    yob = np.linspace(-1200, 200, nyo + 1)
    yyib = np.resize(yib, (nx, nyi + 1))
    dyi = (yib[1] - yib[0]) * 0.49
    yyib += np.random.uniform(-dyi, dyi, yyib.shape)
    yyob = np.resize(yob, (nx, nyo + 1))
    dyo = (yob[1] - yob[0]) * 0.49
    yyob += np.random.uniform(-dyo, dyo, yyob.shape)
    eshapes = np.full((3, 1), nx)

    # input
    u, v = np.mgrid[-3 : 3 : nx * 1j, -3 : 3 : nyi * 1j] - 2
    vari = np.asarray(u**2 + v**2)
    vari[int(nx / 3) : int(2 * nx / 3), int(nyi / 3) : int(2 * nyi / 3)] = np.nan

    # conserv, no extrap
    varoc = interp.cellave1d(vari, yyib, yyob, eshapes, conserv=True, extrap="no")
    sumi = np.nansum(vari * np.diff(yyib, axis=1), axis=1)
    sumo = np.nansum(varoc * np.diff(yyob, axis=1), axis=1)
    np.testing.assert_allclose(sumi, sumo)

    # average, no extrap
    interp.cellave1d(vari, yyib, yyob, eshapes, conserv=0, extrap="no")

    # average, extrap
    varoe = interp.cellave1d(vari, yyib, yyob, eshapes, conserv=False, extrap="both")
    assert not np.isnan(varoe[0]).any()


@pytest.mark.parametrize(
    "x,y,pt,qt",
    [
        (0.0, 0, 0, 0),
        (3, 1, 0, 1),
        (2, 3, 1, 1),
        (-1, 2, 1, 0),
        (1.0, 1.5, 0.5, 0.5),
        (-1, -1, -1, -1),
    ],
)
def test_interp_cell2relloc(x, y, pt, qt):

    x1, y1 = 0.0, 0.0
    x2, y2 = 3.0, 1.0
    x3, y3 = 2.0, 3.0
    x4, y4 = -1.0, 2.0
    p, q = interp.cell2relloc(x1, x2, x3, x4, y1, y2, y3, y4, x, y)
    assert p == pt
    assert q == qt


@functools.lru_cache()
def get_grid2locs_coords(nex=4, nexz=2, nxi=7, nyi=6, nzi=5, nti=4, no=10):

    np.random.seed(0)

    tti, zzi, yyi, xxi = np.mgrid[
        0 : nti - 1 : nti * 1j,
        0 : nzi - 1 : nzi * 1j,
        0 : nyi - 1 : nyi * 1j,
        0 : nxi - 1 : nxi * 1j,
    ]

    zzi = zzi[None]
    zzi = np.repeat(zzi, nexz, axis=0)

    xyztomin = -0.5
    xo = np.random.uniform(xyztomin, nxi - 1.5, no)
    yo = np.random.uniform(xyztomin, nyi - 1.5, no)
    zo = np.random.uniform(xyztomin, nzi - 1.5, no)
    to = np.random.uniform(xyztomin, nti - 1.5, no)

    tti = round_as_time(tti)
    to = round_as_time(to)

    return xxi, yyi, zzi, tti, xo, yo, to, zo


def test_interp_grid2locs():

    # Multi-dimensional generic coordinates
    nex = 4
    nexz = 2
    nxi = 7
    nyi = 6
    nzi = 5
    nti = 4
    no = 10
    xxi, yyi, zzi, tti, xo, yo, to, zo = get_grid2locs_coords(
        nex=nex, nexz=nexz, nxi=nxi, nyi=nyi, nzi=nzi, nti=nti, no=no
    )

    # Pure 1D axes
    xi = xxi[0, 0, 0:1, :]  # (nyix=1,nxi)
    yi = yyi[0, 0, :, 0:1]  # (nyi,nxiy=1)
    zi = zzi[0:1, 0:1, :, 0:1, 0:1]  # (nexz=1,ntiz=1,nzi,nyiz=1,nxiz=1)
    ti = tti[:, 0, 0, 0]  # (nti)
    vi = vfunc(tti, zzi, yyi, xxi)
    vi = np.resize(vi, (nex,) + vi.shape[1:])  # (nex,nti,nzi,nyi,nxi)
    vo_truth = np.array(vfunc(to, zo, yo, xo))
    vo_interp = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)
    assert vo_interp[0].shape == vo_truth.shape
    vo_truth[np.isnan(vo_interp[0])] = np.nan
    np.testing.assert_allclose(vo_interp[0], vo_truth)

    # Single point in space
    xi = xxi[0, 0, 0:1, :1]  # (nyix=1,nxi)
    yi = yyi[0, 0, :1, 0:1]  # (nyi,nxiy=1)
    zi = zzi[0:1, 0:1, :, 0:1, 0:1]  # (nexz=1,ntiz=1,nzi,nyiz=1,nxiz=1)
    ti = tti[:, 0, 0, 0]  # (nti)
    vi = vfunc(tti, zzi, yyi, xxi)[:, :, :, :1, :1]
    vi = np.resize(vi, (nex,) + vi.shape[1:])  # (nex,nti,nzi,1,1)
    vo_truth = np.array(vfunc(to, zo, yi[0], xi[0]))
    vo_interp = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)
    vo_truth[np.isnan(vo_interp[0])] = np.nan
    np.testing.assert_allclose(vo_interp[0], vo_truth)

    # Constant time
    xi = xxi[0, 0, 0:1, :]  # (nyix=1,nxi)
    yi = yyi[0, 0, :, 0:1]  # (nyi,nxiy=1)
    zi = zzi[0:1, 0:1, :, 0:1, 0:1]  # (ntiz=1,nzi,nyiz=1,nxiz=1)
    ti = tti[:1, 0, 0, 0]  # (1)
    vi = vfunc(tti, zzi, yyi, xxi)[:, :1, :, :, :]
    vi = np.resize(vi, (nex,) + vi.shape[1:])  # (nex,1,nzi,nyi,nxi)
    vo_truth = vfunc(ti, zo, yo, xo)
    vo_interp = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)
    vo_truth[np.isnan(vo_interp[0])] = np.nan
    np.testing.assert_allclose(vo_interp[0], vo_truth)

    # Variable depth with 1D X/Y + T
    xi = xxi[0, 0, 0:1, :]  # (nyix=1,nxi)
    yi = yyi[0, 0, :, 0:1]  # (nyi,nxiy=1)
    zi = zzi[:, :, :, :, :]  # (nexz,ntiz=nti,nzi,nyiz=nyi,nxiz=nxi)
    ti = tti[:, 0, 0, 0]  # (nti)
    vi = vfunc(tti, zzi, yyi, xxi)
    vi = np.resize(vi, (nex,) + vi.shape[1:])  # (nex,nti,nzi,nyi,nxi)
    vo_truth = vfunc(to, zo, yo, xo)
    vo_interp = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)
    vo_truth[np.isnan(vo_interp[0])] = np.nan
    np.testing.assert_allclose(vo_interp[0], vo_truth)

    # 2D X/Y with no other axes (pure curvilinear)
    xi = xxi[0, 0]  # (nyix=nyi,nxi)
    yi = yyi[0, 0]  # (nyi,nxiy=nxi)
    zi = zzi[0:1, 0:1, 0:1, 0:1, 0:1]  # (nexz=1,ntiz=1,1,nyiz=1,nxiz=1)
    ti = tti[:1, 0, 0, 0]  # (1)
    vi = vfunc(tti, zzi, yyi, xxi)[:, :1, :1, :, :]
    vi = np.resize(vi, (nex,) + vi.shape[1:])  # (nex,1,1,nyi,nxi)
    vo_interp = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)
    vo_interp_rect = interp.grid2locs(xi[:1], yi[:, :1], zi, ti, vi, xo, yo, zo, to)
    vo_truth = vfunc(ti, zi.ravel()[0], yo, xo)
    vo_truth[np.isnan(vo_interp[0])] = np.nan
    np.testing.assert_allclose(vo_interp[0], vo_truth)
    vo_truth = vfunc(ti, zi.ravel()[0], yo, xo)
    vo_truth[np.isnan(vo_interp_rect[0])] = np.nan
    np.testing.assert_allclose(vo_interp_rect[0], vo_truth)

    # Same coordinates
    xi = xxi[0, 0, 0:1, :]  # (nyix=1,nxi)
    yi = yyi[0, 0, :, 0:1]  # (nyi,nxiy=1)
    zi = zzi[0:1, 0:1, :, 0:1, 0:1]  # (nexz=1,ntiz=1,nzi,nyiz=1,nxiz=1)
    ti = tti[:, 0, 0, 0]  # (nti)
    vi = vfunc(tti, zzi, yyi, xxi)
    vi = np.resize(vi, (nex,) + vi.shape[1:])  # (nex,nti,nzi,nyi,nxi)
    tzyxo = np.meshgrid(ti, zi, yi, xi, indexing='ij')
    xo = tzyxo[3].ravel()
    yo = tzyxo[2].ravel()
    zo = tzyxo[1].ravel()
    to = tzyxo[0].ravel()
    vo_truth = vfunc(to, zo, yo, xo)
    vo_interp = interp.grid2locs(xi, yi, zi, ti, vi, xo, yo, zo, to)
    vo_truth[np.isnan(vo_interp[0])] = np.nan
    np.testing.assert_allclose(vo_interp[0], vo_truth)


def test_interp_isoslice():

    depth = np.linspace(-50, 0.0, 6)
    values = np.linspace(10, 20.0, 6)
    isoval = 15.0

    isodepth = interp.isoslice(depth, values, isoval, False)
    assert isodepth == -25.0
    isodepth = interp.isoslice(depth, values, isoval, True)
    assert isodepth == -25.0

    depth = np.resize(depth, (2,) + depth.shape)
    isodepth = interp.isoslice(depth, values, isoval, False)
    np.testing.assert_allclose(isodepth, [-25.0, -25.0])



#!/usr/bin/env python
# coding: utf-8
"""
Compare Mercator to ARGO
========================

This notebook, we show:

* how to easily rename variables and coordinates in an ARGO and a Mercator datasets,
* how to plot a T-S diagram,
* how to stack geographical coordinates,
* how to compute a geographical distance between two datasets with coordinates,
* how to interpolate a vertical profile with different methods.
"""


# %%
# Initialisations
# ---------------
import xarray as xr
import matplotlib as mpl
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import gsw
import cmocean

import xoa
import xoa.grid as xgrid
import xoa.regrid as xregrid
import xoa.cf as xcf
import xoa.coords as xcoords
import xoa.geo as xgeo
import xoa.plot as xplot

mpl.rc("axes", grid=True)
xr.set_options(display_style="text")

# %%
# Register the :ref:`xoa <accessors>` accessors :
xoa.register_accessors()

# %%
# Register the Mercator and ARGO naming specifications
xcf.register_cf_specs(xoa.get_data_sample("mercator.cfg"))
xcf.register_cf_specs(xoa.get_data_sample("argo.cfg"))

# %%
# Here is what these CF specifications contain
with open(xoa.get_data_sample("mercator.cfg")) as f:
    print(f.read())
with open(xoa.get_data_sample("argo.cfg")) as f:
    print(f.read())

# %%
# Read and decode the datasets
# ----------------------------

# %%
# ARGO profiles
# ~~~~~~~~~~~~~
#
# The ARGO data were downloaded with the `argopy <https://argopy.readthedocs.io/en/latest/>`_
# package with a piece of code close to the following::
#
#     from argopy import DataFetcher
#     adf = DataFetcher().float(7900573)
#     ds = adf.to_xarray().argo.point2profile().isel(N_PROF=slice(-10, None))
#     ds.to_netcdf("argo-7900573.nc")
#
# Here is the web page of this float: https://fleetmonitoring.euro-argo.eu/float/7900573
ds_argo = xoa.open_data_sample("argo-7900573.nc")

# %%
# We decode the names to make them more generic thanks to the CF specifications
# and select only the variables of interest.
ds_argo = ds_argo.xoa.decode()[["temp", "sal", "pres"]]

# %%
# We compute depths with the :mod:`gsw` package and assign them
# as coordinates of the ARGO dataset.
alat = ds_argo.lat.broadcast_like(ds_argo.pres)
adepth = gsw.z_from_p(ds_argo.pres, alat)
adepth = adepth.where(adepth.notnull(), adepth.min())
adepth.attrs.update({"long_name": "Depth", "units": "m"})
ds_argo = ds_argo.assign_coords(depth=adepth)
ds_argo = ds_argo.isel(level=slice(None, None, -1))
ds_argo = ds_argo.set_index(N_PROF="time").rename(N_PROF="time")

# %%
# Quick plot of the salinity that highligths the mediterranean water.
fig, axs = plt.subplots(ncols=2, figsize=(14, 6))
ds_argo.sal.plot.contourf("time", "depth", cmap="cmo.haline", levels=20, ax=axs[0])
ds_argo.sal.plot.contour("time", "depth", levels=[35.65], linewidths=1, colors="k", ax=axs[0])
xplot.plot_ts(
    ds_argo.temp,
    ds_argo.sal,
    potential=False,
    scatter_c=ds_argo.depth,
    scatter_cmap="cmo.deep",
    axes=axs[1],
)
axs[1].axvline(x=35.65, ls="--", color="k");

# %%
# Here we select the last ARGO profile as a reference profile
# since it better samples de mediterranean water vein.
ds_argo_prof = ds_argo.isel(time=-1).squeeze(drop=True)

# %%
# Mercator
# ~~~~~~~~
#
# The Mercator data come from the IBI configuration and were downloaded
# from the CMEMS site:
# https://resources.marine.copernicus.eu/product-detail/IBI_ANALYSISFORECAST_PHY_005_001/INFORMATION
# The dataset has been undersampled by a factor 3 in longitude and latitude
# to limit disk usage.
ds_merc = xoa.open_data_sample("ibi-argo-7900573.nc").xoa.decode()
ds_merc = ds_merc.isel(depth=slice(None, None, -1))
ds_merc = ds_merc.assign_coords(depth=-ds_merc.depth)
print(ds_merc)

# %%
# By analogy with the ensemble data assimilation,
# this block of data can be viewed as a ensemble of profiles that emulates
# the uncertainties of the model, so we stack geographical coordinates in a `member` dimension
# thanks to :func:`~xoa.coords.geo_stack`.
ds_merc_ens_rect = xcoords.geo_stack(ds_merc, "member")
print(ds_merc_ens_rect)

# %%
# This is a rectangular selection and we find it more appropriate to retain
# only profiles that are in a given radius of proximity, that is chosen
# close to the surface radius of deformation.
# We compute the distance from each model profile to the selected ARGO profile
# with :func:`~xoa.geo.get_distances`.
radius = 35e3  # 35 km
dist_merc2argo = (
    xgeo.get_distances(ds_merc_ens_rect, ds_argo_prof)
    .squeeze()
    .rename(npts0="member")
    .assign_coords(member=ds_merc_ens_rect.member)
)
ds_merc_ens = ds_merc_ens_rect.where(dist_merc2argo < radius, drop=True)


# %%
# Compare Mercator and ARGO
# -------------------------
#

# %%
# Let's plot the situation.
pmerc = ccrs.Mercator()
pcar = ccrs.PlateCarree()
fig, ax = plt.subplots(subplot_kw={"projection": pmerc}, figsize=(8, 8))
ax.set_extent(xgeo.get_extent(ds_merc, margin=1, square=True))
ax.gridlines(draw_labels=True)
ax.add_wms("https://ows.emodnet-bathymetry.eu/wms", "emodnet:mean_atlas_land", alpha=0.5)
kws = dict(c="C3", s=15, marker="s", transform=pcar)
ax.scatter(ds_merc_ens_rect.lon, ds_merc_ens_rect.lat, label="Rejected", alpha=0.15, **kws)
ax.scatter(ds_merc_ens.lon, ds_merc_ens.lat, label='Mercator', **kws)
ax.scatter(ds_argo_prof.lon, ds_argo_prof.lat, s=100, c="C2", transform=pcar, label='ARGO')
plt.legend();

# %%
# We used the :func:`xoa.geo.get_extent` to compute a square geographical extent based
# on the coordinates of a dataset with an added margin.

# %%
# Now interpolate Mercator onto ARGO time and position for both the ensemble and
# the reference profile. All interpolations can be performed
# with the :meth:`~xarray.Dataset.interp` method since we are dealing with
# axis (1D) coordinates.
ds_merc_ens = ds_merc_ens.interp(time=ds_argo_prof.time.values)
ds_merc_ens_std = ds_merc_ens.std("member")
ds_merc_prof = ds_merc.interp(time=ds_argo_prof.time, lon=ds_argo_prof.lon, lat=ds_argo_prof.lat)

# %%
# We plot now the model, its simulated uncertainty and the observations.
plt.figure(figsize=(4, 7))
plt.fill_betweenx(
    ds_merc_prof.sal.depth,
    ds_merc_prof.sal - 1.96 * ds_merc_ens_std.sal,
    ds_merc_prof.sal + 1.96 * ds_merc_ens_std.sal,
    fc="C1",
    alpha=0.2,
)
ds_merc_prof.sal.plot(y="depth", color="C1", label="Mercator")
plt.plot(ds_merc_prof.sal.values, ds_merc_prof.depth, "o-", color="C1")
ds_argo_prof.sal.plot(y="depth", label="ARGO", color="C0")
plt.title("Uncertain profile")
plt.legend()
xplot.plot_double_minimap(ds_argo_prof, regional_ax="left", global_ax=(0.88, 0.88, 0.11),);

# %%
# Therefore, if we accept an uncertainty in the positioning of the ocean
# structure in the model, it differences with (assimilated) observations
# become acceptables.

# %%
# As you can see on the previous figures, the ARGO profiles come with
# a very high vertical resolution at all depths, whereas the model
# comes with a crude vertical resolution far from the surface.
# Each model thermodynamical model value is representative of the
# cube that defines its XYZ cell.
# This means that despite the model does simulate the small scale variability,
# the cell average may be accurate.
# So we propose here to perform comparisons in the model
# and the observational spaces. We use the function :func:`xoa.regrid.regrid1d`
# that supports the linear (:attr:`~xoa.regrid.regrid1d_methods.linear` method) interpolation
# to higher resolutions and the cell averaging
# (:attr:`~xoa.regrid.regrid1d_methods.cellave` method) to lower resolutions.
sal_merc_ens_o = xregrid.regrid1d(ds_merc_ens.sal, ds_argo_prof.depth, method="linear")
sal_merc_prof_o = xregrid.regrid1d(ds_merc_prof.sal, ds_argo_prof.depth, method="linear")
sal_argo_prof_m = xregrid.regrid1d(ds_argo_prof.sal, ds_merc_prof.depth, method="cellave")
merc_prof_edges = xgrid.get_edges(ds_merc_prof.depth, "depth")

# %%
# Let's plot them.
fig, axes = plt.subplots(ncols=2, sharex=True, sharey=True, figsize=(10, 5))
sal_merc_prof_o.plot(y="depth", color="C1", label="Mercator", ax=axes[0])
ds_argo_prof.sal.plot(y="depth", label="ARGO", color="C0", ax=axes[0])
axes[0].set_title("On ARGO depths")
axis = axes[0].axis()
ds_merc_prof.sal.plot(y="depth", color="C1", label="Mercator", ax=axes[1])
plt.stairs(ds_merc_prof.sal, merc_prof_edges, orientation="horizontal", color="C1", lw=0.5)
sal_argo_prof_m.plot(y="depth", label="ARGO", color="C0", ax=axes[1])
plt.stairs(sal_argo_prof_m, merc_prof_edges, orientation="horizontal", color="C0", lw=0.5)
axes[1].set_title("On Mercator depths")
axes[1].axis(axis);

# %%
# These plots show that the model is too haline, except at the
# mediterranean water depth where the model is very close to the observations
# once an appropriate vertical regridding is performed.
# However, the vertical resolution is not sufficient enough to simulate
# the maximum of salinity.



# -*- coding: utf-8 -*-
"""
Plotting utilities

Filters are adapted from https://matplotlib.org/stable/gallery/misc/demo_agg_filter.html?highlight=agg%20filter
and http://vacumm.github.io/vacumm/library/misc.core_plot.html

"""
# Copyright 2020-2022 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.collections as mcollections
import matplotlib.artist as martist
import matplotlib.text as mtext
import matplotlib.patheffects as mpatheffects
import xarray as xr
import cartopy.crs as ccrs
import cartopy.feature as cfeature

from .__init__ import xoa_warn
from . import misc as xmisc
from . import geo as xgeo
from . import cf as xcf
from . import coords as xcoords
from . import dyn

# %% Special functions


def plot_flow(
    u,
    v,
    duration=None,
    step=None,
    particles=2000,
    axes=None,
    alpha=(0.2, 1),
    linewidth=0.3,
    color="k",
    autolim=None,
    **kwargs,
):
    """Plot currents as a windy-like plot with random little lagrangian tracks

    Parameters
    ----------
    u: xarray.DataArray
        Gridded zonal velocity
    v: xarray.DataArray
        Gridded meridional velocity
    duration: int, numpy.timedelta64
        Total integration time in seconds
    step: int, numpy.timedelta64
        Integration step in seconds
    particles: int, xarray.Dataset, tuple
        Either a number of particles or a dataset of initial positions
        with longitude and latitude coordinates
    axes: matplotlib.axes.Axes
        The axes instance
    alpha: float, tuple
        Alpha transparency. If a tuple, apply a linear alpha ramp to the track
        from its start to its end.
    linewidth: float, tuple
        Linewidth of the track. If a tuple, apply a linear linewidth ramp to the track
        from its start to its end.
    color:
        Single color for the track.
    autolim: None, bool
        Wether to auto-update the data limits.
        A value of None sets autolim to True if "axes" is not provided,
        else to None. See :meth:`matplotlib.axes.Axes.add_collection`.

    Return
    ------
    Return
    ------
    dict
        With the following keys:
        `axes`, `linecollection`, `step`, `duration`.
        `linecollection` refere to the :class:`matplotlib.collections.LineCollection` instance.
        The duration and step are also stored in the output.

    Example
    ------
    .. ipython:: python

        @suppress
        import numpy as np, xarray as xr
        @suppress
        from xoa.plot import plot_flow
        # Setup data
        x = np.linspace(0, 2*np.pi, 20)
        y = np.linspace(0, 2*np.pi, 20)
        X, Y = np.meshgrid(x,y)
        U = np.sin(X) * np.cos(Y)
        V = -np.cos(X) * np.sin(Y)

        # As a dataset
        ds = xr.Dataset(
            {"u": (('lat', 'lon'), U), "v": (('lat', 'lon'), V)},
            coords={"lat": ("lat", y), "lon": ("lon", x)})

        # Plot
        @savefig api.plot.plot_flow.png
        plot_flow(ds["u"], ds["v"])

    See also
    --------
    xoa.dyn.flow2d
    matplotlib.axes.Axes.add_collection
    """
    # Infer parameters
    if duration is None or step is None:

        # Default duration based on a the fraction of the area crossing time
        if duration is None:
            xmin, xmax, ymin, ymax = xgeo.get_extent(u)
            dist = xgeo.haversine(xmin, ymin, xmax, ymax)
            speed = np.sqrt(np.abs(u.values).mean() ** 2 + np.abs(v.values).mean() ** 2)
            duration = dist / speed / 30
        elif isinstance(duration, np.timedelta64):
            duration /= np.timedelta64(1, "s")

        # Step
        if step is None:
            step = duration / 10
    if isinstance(duration, np.timedelta64):
        duration /= np.timedelta64(1, "s")
    if isinstance(step, np.timedelta64):
        step /= np.timedelta64(1, "s")

    # Compute the flow
    flow = dyn.flow2d(u, v, particles, duration, step)
    tx = flow.lon.values
    ty = flow.lat.values

    # Get the distance from start for each track
    ramp = not np.isscalar(alpha) or not np.isscalar(linewidth)
    if ramp:
        dists = xgeo.haversine(tx[:-1], ty[:-1], tx[1:], ty[1:])
        cdists = np.cumsum(dists, axis=0)
        cdists /= np.nanmax(cdists)
        cdists[np.isnan(cdists)] = 0

    # Plot specs
    color = mcolors.to_rgb(color)
    segments = []
    linewidths = []
    colors = []
    for j in range(tx.shape[0] - 1):
        for i in range(tx.shape[1]):
            segments.append(((tx[j, i], ty[j, i]), (tx[j + 1, i], ty[j + 1, i])))
            if not np.isscalar(alpha):
                colors.append(color + (alpha[0] + cdists[j, i] * (alpha[-1] - alpha[0]),))
            else:
                colors.append(color + (alpha,))
            if not np.isscalar(linewidth):
                linewidths.append(
                    linewidth[0] + cdists[j, i] * (linewidth[-1] - linewidth[0]),
                )
            else:
                linewidths.append(linewidth)

    # Plot
    kwargs.setdefault("colors", colors)
    kwargs.setdefault("linewidths", linewidths)
    lc = mcollections.LineCollection(segments, **kwargs)
    axes = kwargs.get("ax", axes)
    newaxes = axes is None
    if newaxes:
        axes = plt.gca()
    if autolim is None:
        autolim = newaxes
    axes.add_collection(lc, autolim=autolim)

    return {"axes": axes, "step": step, "duration": duration, "linecollection": lc}


def plot_ts(
    temp,
    sal,
    dens=True,
    pres=None,
    potential=None,
    axes=None,
    scatter_kwargs=None,
    contour_kwargs=None,
    clabel=True,
    clabel_kwargs=None,
    colorbar=None,
    colorbar_kwargs=None,
    **kwargs,
):
    """Plot a TS diagram

    A TS diagram is a scatter plot with absolute salinity as X axis
    and potential temperature as Y axis.
    The density is generally added as background contours.

    Parameters
    ----------
    temp: xarray.DataArray
        Temperature. If not potential temperature, please set `potential=True`.
    sal: xarray.DataArray
        Salinity
    dens: bool
        Add contours of density.
        The density is computed with function :func:`gsw.density.sigma0`.
    pres: xarray.DataArray, None
        Pressure to compute potential temperature.
    potential: bool, None
        Is the temperature potential? If None, infer from attributes.
    clabel: bool
        Add labels to density contours
    clabel_kwargs: dict, None
        Parameters that are passed to :func:`~matplotlib.pyplot.clabel`.
    colorbar: bool, None
        Should we add the colorbar? If None, check if scatter plot color is a data array.
    colorbar_kwargs: dict, None
        Parameters that are passed to :func:`~matplotlib.pyplot.colorbar`.
    contour_kwargs: dict, None
        Parameters that are passed to :func:`~matplotlib.pyplot.contour`.
    axes: None
        Matplotlib axes instance
    kwargs: dict
        Extra parameters are filtered by :func:`xoa.misc.dict_filter`
        and passed to the plot functions.

    See also
    --------
    :mod:`gsw.density`
    :mod:`gsw.conversions`

    Return
    ------
    dict
        With the following keys, depending on what is plotted:
        `axes`, `scatter`, `colorbar`, `contour`, `clabel`.


    Example
    -------
    .. ipython:: python

        @suppress
        import numpy as np, xarray as xr, xoa, xoa.coords, cmocean
        @suppress
        from xoa.plot import plot_ts

        # Register the main xoa accessor
        xoa.register_accessors()

        # Load the CROCO meridional section
        ds = xoa.open_data_sample("croco.south-africa.meridional.nc")
        ds = ds.isel(eta_rho=slice(40))
        temp = ds.xoa.get('temp')     # requests are based...
        sal = ds.xoa.get('sal')       # ...on the generic name
        depth = ds.xoa.get_depth(ds)  # or xoa.coords.get_depth(ds)

        # Plot
        @savefig api.plot.plot_ts.png
        plot_ts(temp, sal, potential=True, scatter_c=depth, contour_linewidths=0.2, clabel_fontsize=8)

    """

    # Potential temperature
    cfspecs = xcf.get_cf_specs(temp)
    # potential = POTENTIAL[potential]
    if potential is None:
        potential = cfspecs.match_data_var(temp, "ptemp")
    if not potential:
        import gsw

        if pres is None:
            lat = xcoords.get_lat(temp)
            depth = xcoords.get_depth(temp)
            lat, depth = xr.broadcast(lat, depth)
            pres = gsw.p_from_z(depth, lat)
        attrs = temp.attrs
        temp = gsw.pt0_from_t(sal, temp, pres)
        temp.attrs.update(attrs)
        cfspecs.format_data_var(temp, cf_name="ptemp", copy=False, replace_attrs=True)

    # Init plot
    axes = kwargs.get("ax", axes)
    if axes is None:
        axes = plt.gca()
    out = {"axes": axes}

    # Scatter plot
    scatter_kwargs = xmisc.dict_filter(
        kwargs, "scatter_", defaults={"s": 10}, **(scatter_kwargs or {})
    )
    out["scatter"] = axes.scatter(sal.values, temp.values, **scatter_kwargs)

    # Colorbar
    if colorbar is None:
        colorbar = "c" in scatter_kwargs and hasattr(scatter_kwargs["c"], "data")
    if colorbar:
        colorbar_kwargs = xmisc.dict_filter(
            kwargs, "colorbar_", defaults={}, **(colorbar_kwargs or {})
        )
        c = scatter_kwargs["c"]
        if "label" not in colorbar_kwargs and hasattr(c, "attrs"):
            label = c.attrs.get("long_name") or c.name
            if label:
                label = label.title()
                units = c.attrs.get("units")
                if units:
                    label = f"{label} [{units}]"
                colorbar_kwargs["label"] = label
        out["colorbar"] = plt.colorbar(out["scatter"], ax=axes, **colorbar_kwargs)

    # Labels
    axes.set_xlabel(sal.attrs.get("long_name", "Salinity").title())
    tlabel = temp.attrs.get("long_name", "Temperature").title()
    tunits = temp.attrs.get("units", "°C")
    axes.set_ylabel(f"{tlabel} [{tunits}]")

    # Density contours
    if dens is not False:

        import gsw

        # Density as sigma0
        (smin, tmin), (smax, tmax) = axes.viewLim.get_points()
        tt = np.linspace(tmin, tmax, 100)
        ss = np.linspace(smin, smax, 100)
        ss, tt = np.meshgrid(ss, tt)
        dd = gsw.sigma0(ss, tt)

        # Contours
        contour_kwargs = xmisc.dict_filter(
            kwargs, "contour_", defaults={"colors": ".3"}, **(contour_kwargs or {})
        )
        out["contour"] = axes.contour(ss, tt, dd, **contour_kwargs)

        # Contour labels
        clabel_kwargs = xmisc.dict_filter(kwargs, "clabel_", defaults={}, **(clabel_kwargs or {}))
        out["clabel"] = axes.clabel(out["contour"], **clabel_kwargs)

    return out


minimap_plot_coords = xmisc.Choices(
    {
        "auto": "infer from coords",
        "box": "rectangle of coords extent",
        "filledbox": "filled rectangle of coords extent",
        "points": "scatter plot",
        "lines": "as lines",
        "center": "a single point at the center",
        "Filled": "filled polygon",
        False: "do not plot",
    },
    parameter="plot_coords",
    description="Type of plot for coordinates",
    aliases={"auto": [True, None]},
)


@minimap_plot_coords.format_function_docstring
def plot_minimap(
    obj,
    ax=[0.9, 0.9, 0.09],
    fig=None,
    extent=1.0,
    min_extent=2.0,
    gridlines=True,
    ocean_color=None,
    land_color=None,
    land_scale="110m",
    plot_coords="auto",
    coords_markersize=10,
    coords_linewidth=2,
    coords_color="tab:red",
    coords_facecolor=0.2,
    **kwargs,
):
    """Plot a small map to show the geographic situation of coordinates

    Parameters
    ----------
    obj: xarray.DataArray, xarray.Dataset, tuple
        Object that contains lon and lat coordinates.
        The object must contains unique and identifiable geographic coordinates
        that can be retrieved xith :func:`xoa.coords.get_lon` and func:`xoa.coords.get_lat`.
        In case of a tuple, it should a couple of `(lon, lat)` :class:`xarray.DataArray`.
    ax: list, cartopy.mpl.geoaxes.GeoAxes
        A matplotlib axes instance or a list that defines the bounding box of the axes to be
        created ``[xmin, ymin, width, height]`` or ``[xmin, ymin, size]`` in figure coordinates.
    fig: figure
        Figure instance
    extent: str, float
        Either ``"global"`` for a global minimap, or the margin added to the coordinates
        bounding box expressed relative to the coordinates extent: a velue of 1.0 means
        a margin equal to the extent.
    min_extent: None, float, (float, float)
        Minimal extent in degrees. See :func:`xoa.geo.get_extent`
    gridlines: bool
        Add gridlines.
    ocean_color: None, color
        A matplotlib color for oceans.
        If `None`, it defaults to ``cartopy.feature.COLORS["ocean"]``.
    land_color: None, color
        A matplotlib color for lands.
        If `None`, it defaults to ``cartopy.feature.COLORS["land"]``.
    {plot_coords}
    coords_markersize: float
        Size of markers in ``"points"`` mode.
    coords_linewidth: float
        Line width in ``"lines"``, `"box"`` and ``"filledbox"`` modes.
    coord_color: color
        Matplotlib color.
    coords_facecolor: color
        Matplotlib face color in ``"filled"`` and ``"filledbox"`` modes.
        If a float, it is interpreted as an alpha transparency that is applied to `coord_color`
        to get the face color.
    **kwargs:
        Parameters matching ``land_<param>`` are passed to the function that plots
        lands.
        Parameters matching ``coords_<param>`` are passed to the function that plots
        coordinates.
        Parameters matching ``gridlines_<param>`` are passed to
        :meth:`cartopy.mpl.geoaxes.GeoAxes.gridlines`.

    Return
    ------
    cartopy.mpl.geoaxes.GeoAxes
        An instance of cartopy geographic axes

    Example
    -------
    .. ipython:: python

        @suppress
        import xarray as xr, numpy as np, matplotlib.pyplot as plt
        @suppress
        from xoa.plot import plot_minimap
        ds = xr.Dataset(coords={{"lon": ("station", [-7, -5, -5]), "lat": ("station", [44, 44, 46])}})
        plt.plot([0, 2], [0, 2])
        @savefig api.plot.plot_minimap.global.png
        plot_minimap(ds, extent="global")
        plt.figure()
        plt.plot([0, 2], [0, 2])
        @savefig api.plot.plot_minimap.regional.png
        plot_minimap(ds, extent=1.2, color="tab:green", gridlines=False, land_color="k")

    See also
    --------
    plot_double_minimap
    xoa.coords.get_lon
    xoa.coords.get_lat
    xoa.geo.get_extent
    """
    # Create map
    pcar = ccrs.PlateCarree()
    if isinstance(obj, tuple):
        lon, lat = obj
    else:
        lon = xcoords.get_lon(obj)
        lat = xcoords.get_lat(obj)
    if fig is None:
        fig = plt.gcf()
    if isinstance(ax, (list, tuple)):
        if ocean_color is None:
            ocean_color = cfeature.COLORS["water"]
        proj = ccrs.NearsidePerspective(
            central_longitude=float(lon.mean()),
            central_latitude=float(lat.mean()),  # satellite_height=50000
        )
        if len(ax) == 3:
            ax = ax + ax[-1:]
        ax = fig.add_axes(ax, projection=proj, facecolor=ocean_color)
        ax.spines["geo"].set_linewidth(0.2)
    if gridlines:
        kwgl = xmisc.dict_filter(kwargs, "gridlines_")
        ax.gridlines(**kwgl)
    if land_scale is None:
        land_scale = "50m" if extent == "global" else "110m"
    if extent == "global":
        ax.set_global()
    else:
        extent = np.array(xgeo.get_extent(obj, square=True, margin=extent, min_extent=min_extent))
        ax.set_extent(extent, pcar)

    # Add land
    kwland = xmisc.dict_filter(kwargs, "land_")
    if land_color is None:
        land_color = cfeature.COLORS["land"]
    kwland["facecolor"] = land_color
    ax.add_feature(cfeature.LAND.with_scale(land_scale), **kwland)

    # Add coordinates
    plot_coords = minimap_plot_coords[plot_coords]
    if plot_coords is not False:
        kwcoords = xmisc.dict_filter(kwargs, "coords_")
        kwcoords.update(transform=pcar)
        if plot_coords in ("box", "filledbox"):
            bbox = xgeo.get_extent((lon, lat))
            lon = [bbox[0], bbox[1], bbox[1], bbox[0], bbox[0]]
            lat = [bbox[2], bbox[2], bbox[3], bbox[3], bbox[2]]
            plot_coords = "lines" if plot_coords == "box" else "filled"
        elif plot_coords == "center":
            lon = [lon.mean()]
            lat = [lat.mean()]
            plot_coords = "points"
        if plot_coords is None:
            plot_coords = "auto"
        if plot_coords == "auto":
            if lon.dims == lat.dims and lon.ndim == 1 and lon.size > 0:
                plot_coords = "lines"
            else:
                lon, lat = xr.broadcast(lon, lat)
                plot_coords = "points"
        if plot_coords == "lines":
            kwcoords.update(linewidth=coords_linewidth, color=coords_color)
            ax.plot(lon, lat, **kwcoords)
        elif plot_coords == "filled":
            if isinstance(coords_facecolor, float) and coords_facecolor <= 1:
                fc_alpha = coords_facecolor
                coords_facecolor = mcolors.to_rgba(coords_color)
                fc_alpha *= coords_facecolor[-1]
                coords_facecolor = coords_facecolor[:3] + (fc_alpha,)
            kwcoords.update(
                linewidth=coords_linewidth, color=coords_color, facecolor=coords_facecolor
            )
            ax.fill(lon, lat, **kwcoords)
        else:
            kwcoords.update(s=coords_markersize, c=coords_color)
            ax.scatter(lon, lat, **kwcoords)

    return ax


def plot_double_minimap(obj, regional_ax="below", **kwargs):
    """Plot a global minimap and a regional minimap

    It consists in two calls to :func:`plot_minimap` with the first one whith a "global" extent.
    By default, the coordinates are plotted as single point on the global minimap, and the
    regional minimap is placed below the global one.

    Parameters
    ----------
    regional_ax: axes, list, str
        If a string, it should be one of ``"below"``, ``"above"``, ``"left"`` or ``"right"``
        and it is interpreted as a relative position of the regional minimap with respect
        to the global one.
    **kwargs:
        Parameters matching ``global_<param>`` are passed to the global minimap while
        parameters matching ``regional_<param>`` are passed to the regional minimap.
        All other parameters are passed to both minimaps.

    Returns
    -------
    cartopy.mpl.geoaxes.GeoAxes, cartopy.mpl.geoaxes.GeoAxes
        A tuple of `(global_ax, regiona_ax)`

    Example
    -------
    .. ipython:: python

        @suppress
        import xarray as xr, numpy as np, matplotlib.pyplot as plt
        @suppress
        from xoa.plot import plot_double_minimap
        ds = xr.Dataset(coords={"lon": ("station", [-7, -5, -5]), "lat": ("station", [44, 44, 46])})
        plt.plot([0, 2], [0, 2])
        @savefig api.plot.plot_double_minimap.png
        plot_double_minimap(ds)

    See also
    --------
    plot_minimap
    """

    # Filter keywords
    kwglobal = xmisc.dict_filter(kwargs, "global_")
    kwregional = xmisc.dict_filter(kwargs, "regional_")

    # Global minimap
    kw = kwargs.copy()
    kwglobal.update(extent="global")
    kwglobal.setdefault("plot_coords", "center")
    kw.update(kwglobal)
    global_ax = plot_minimap(obj, **kw)

    # Regional minimap
    if isinstance(regional_ax, str):
        bb = global_ax.bbox.transformed(global_ax.figure.transFigure.inverted())
        if regional_ax == "right":
            regional_ax = [bb.xmin + bb.width * 1.1, bb.ymin, bb.width, bb.height]
        elif regional_ax == "above":
            regional_ax = [bb.xmin, bb.ymax + bb.height * 0.1, bb.width, bb.height]
        elif regional_ax == "left":
            regional_ax = [bb.xmin - bb.width * 1.1, bb.ymin, bb.width, bb.height]
        else:
            regional_ax = [bb.xmin, bb.ymin - bb.height * 1.1, bb.width, bb.height]
    kw = kwargs.copy()
    kwregional.update(ax=regional_ax)
    kw.update(kwregional)
    regional_ax = plot_minimap(obj, **kw)

    return global_ax, regional_ax


# %% Filters


def _smooth2d_(A, sigma):
    from scipy.ndimage import gaussian_filter

    return gaussian_filter(A, sigma, truncate=3)


class _BaseFilter_(object):
    def prepare_image(self, src_image, dpi, pad):
        ny, nx, depth = src_image.shape
        padded_src = np.zeros([pad * 2 + ny, pad * 2 + nx, depth], dtype="d")
        padded_src[pad:-pad, pad:-pad, :] = src_image[:, :, :]
        return padded_src

    def get_pad(self, dpi):
        return 0

    def __call__(self, im, dpi):
        pad = self.get_pad(dpi)
        padded_src = self.prepare_image(im, dpi, pad)
        tgt_image = self.process_image(padded_src, dpi)
        return tgt_image, -pad, -pad


class OffsetFilter(_BaseFilter_):
    def __init__(self, offsets=None):
        if offsets is None:
            self.offsets = (0, 0)
        else:
            self.offsets = offsets

    def get_pad(self, dpi):
        return int(max(*self.offsets) / 72.0 * dpi)

    def process_image(self, padded_src, dpi):
        ox, oy = self.offsets
        a1 = np.roll(padded_src, int(ox / 72.0 * dpi), axis=1)
        a2 = np.roll(a1, -int(oy / 72.0 * dpi), axis=0)
        return a2


class GaussianFilter(_BaseFilter_):
    """Gaussian filter"""

    def __init__(self, sigma, alpha=0.5, color=None):
        self.sigma = sigma
        self.alpha = alpha
        if color is None:
            self.color = (0, 0, 0)
        else:
            self.color = color

    def get_pad(self, dpi):
        return int(self.sigma * 3 / 72.0 * dpi)

    def process_image(self, padded_src, dpi):
        # offsetx, offsety = int(self.offsets[0]), int(self.offsets[1])
        tgt_image = np.zeros_like(padded_src)
        aa = _smooth2d_(padded_src[:, :, -1] * self.alpha, self.sigma / 72.0 * dpi)
        tgt_image[:, :, -1] = aa
        tgt_image[:, :, :-1] = self.color
        return tgt_image


class DropShadowFilter(_BaseFilter_):
    """Create a drop shadow"""

    def __init__(self, width, alpha=0.3, color=None, offsets=None):
        self.gauss_filter = GaussianFilter(width / 3, alpha, color)
        self.offset_filter = OffsetFilter(offsets)

    def get_pad(self, dpi):
        return max(self.gauss_filter.get_pad(dpi), self.offset_filter.get_pad(dpi))

    def process_image(self, padded_src, dpi):
        t1 = self.gauss_filter.process_image(padded_src, dpi)
        t2 = self.offset_filter.process_image(t1, dpi)
        return t2


class GrowFilter(_BaseFilter_):
    "Enlarge the area"

    def __init__(self, pixels, color=None, alpha=1.0):
        self.pixels = pixels
        if color is None:
            self.color = (1, 1, 1)
        else:
            self.color = color
        self.alpha = alpha

    def __call__(self, im, dpi):
        pad = self.pixels
        ny, nx, depth = im.shape
        new_im = np.empty([pad * 2 + ny, pad * 2 + nx, depth], dtype="d")
        alpha = new_im[:, :, 3]
        alpha.fill(0)
        alpha[pad:-pad, pad:-pad] = im[:, :, -1]
        alpha2 = np.clip(_smooth2d_(alpha, self.pixels / 72.0 * dpi) * 5, 0, 1) * self.alpha
        new_im[:, :, -1] = alpha2
        new_im[:, :, :-1] = self.color
        offsetx, offsety = -pad, -pad

        return new_im, offsetx, offsety


class LightFilter(_BaseFilter_):
    """Apply a light filter"""

    def __init__(self, sigma, fraction=0.5, **kwargs):
        self.gauss_filter = GaussianFilter(sigma / 3, alpha=1)
        self.light_source = mcolors.LightSource(**kwargs)
        self.fraction = fraction

    def get_pad(self, dpi):
        return self.gauss_filter.get_pad(dpi)

    def process_image(self, padded_src, dpi):
        t1 = self.gauss_filter.process_image(padded_src, dpi)
        elevation = t1[:, :, 3]
        rgb = padded_src[:, :, :3]
        rgb2 = self.light_source.shade_rgb(rgb, elevation, fraction=self.fraction)
        tgt = np.empty_like(padded_src)
        tgt[:, :, :3] = rgb2
        tgt[:, :, 3] = padded_src[:, :, 3]

        return tgt


class FilteredArtistList(martist.Artist):
    """
    A simple container to draw filtered artist.
    """

    def __init__(self, artist_list, filter):
        self._artist_list = artist_list
        self._filter = filter
        super().__init__()

    def draw(self, renderer):
        renderer.start_rasterizing()
        if hasattr(renderer, 'start_filter'):
            renderer.start_filter()
        for a in self._artist_list:
            if hasattr(a, 'draw'):
                a.draw(renderer)
        renderer.stop_filter(self._filter)
        renderer.stop_rasterizing()


def add_agg_filter(objs, filter, zorder=None, ax=None, add=True):
    """Add a filtered version of objects to plot

    Parameters
    ----------

    objs: :class:`matplotlib.artist.Artist`
        Plotted objects.
    filter: :class:`BaseFilter`
    zorder: optional
        zorder (else guess from ``objs``).
    ax: optional, :class:`matplotlib.axes.Axes`
    """
    # Input
    if not isinstance(objs, (list, tuple)):
        objs = [objs]
    elif len(objs) == 0:
        return []

    # Filter
    if ax is None:
        ax = plt.gca()
    shadows = FilteredArtistList(objs, filter)
    if hasattr(add, 'add_artist'):
        add.add_artist(shadows)
    elif add:
        ax.add_artist(shadows)

    # Text
    for t in objs:
        if isinstance(t, mtext.Text):
            t.set_path_effects([mpatheffects.Normal()])

    # Adjust zorder
    if zorder is None or zorder is True:
        same = zorder is True
        if hasattr(objs, 'get_zorder'):
            zorder = objs.get_zorder()
        else:
            zorder = objs[0].get_zorder()
        if not same:
            zorder -= 0.1
    if zorder is not False:
        shadows.set_zorder(zorder)

    return shadows


def add_shadow(
    objs, width=3, xoffset=2, yoffset=-2, alpha=0.5, color='k', zorder=None, ax=None, add=True
):
    """Add a drop-shadow to objects

    Parameters
    ----------
    objs: :class:`matplotlib.artist.Artist`
        Plotted objects.
    width: optional
        Width of the gaussian filter in points.
    xoffset: optional
        Shadow offset along X in points.
    yoffset: optional
        Shadow offset along Y in points.
    color: optional
        Color of the shadow.
    zorder: optional
        zorder (else guess from ``objs``).
    ax: optional, :class:`matplotlib.axes.Axes`

    Inspired from http://matplotlib.sourceforge.net/examples/pylab_examples/demo_agg_filter.html .
    """
    if color is not None:
        color = mcolors.ColorConverter().to_rgb(color)
    try:
        gauss = DropShadowFilter(width, offsets=(xoffset, yoffset), alpha=alpha, color=color)
        return add_agg_filter(objs, gauss, zorder=zorder, ax=ax, add=add)
    except:
        xoa_warn('Cannot plot shadows using agg filters')


def add_glow(objs, width=3, zorder=None, color='w', ax=None, alpha=1.0, add=True):
    """Add a glow effect to text

    Parameters
    ----------
    objs: :class:`matplotlib.artist.Artist`
        Plotted objects.
    width: optional
        Width of the gaussian filter in points.
    color: optional
        Color of the shadow.
    zorder: optional
        zorder (else guess from ``objs``).
    ax: optional, :class:`matplotlib.axes.Axes`

    Inspired from http://matplotlib.sourceforge.net/examples/pylab_examples/demo_agg_filter.html .
    """
    if color is not None:
        color = mcolors.ColorConverter().to_rgb(color)
    try:
        white_glows = GrowFilter(width, color=color, alpha=alpha)
        return add_agg_filter(objs, white_glows, zorder=zorder, ax=ax, add=add)
    except:
        xoa_warn('Cannot add glow effect using agg filters')


def add_lightshading(objs, width=7, fraction=0.5, zorder=None, ax=None, add=True, **kwargs):
    """Add a light shading effect to objects

    Parameters
    ----------
    objs: :class:`matplotlib.artist.Artist`
        Plotted objects.
    width: optional
        Width of the gaussian filter in points.
    fraction: optional
        Unknown.
    zorder: optional
        zorder (else guess from ``objs``).
    ax: optional, :class:`matplotlib.axes.Axes`
    **kwargs
        Extra keywords are passed to :class:`matplotlib.colors.LightSource`

    Inspired from http://matplotlib.sourceforge.net/examples/pylab_examples/demo_agg_filter.html .
    """
    if zorder is None:
        zorder = True
    try:
        lf = LightFilter(width, fraction=fraction, **kwargs)
        return add_agg_filter(objs, lf, zorder=zorder, add=add, ax=ax)
    except:
        xoa_warn('Cannot add light shading effect using agg filters')



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.filter` module
"""

import pytest
import numpy as np
import xarray as xr

import xoa.filter as xfilter

np.random.seed(0)
da1d = xr.DataArray(np.random.normal(size=10), dims=('nt',))
da2d = xr.DataArray(np.random.normal(size=(50, 50)), dims=('ny', 'nx'))
da3d = xr.DataArray(np.random.normal(size=(30, 50, 50)), dims=('nt', 'ny', 'nx'))


def test_filter_generate_orthogonal_kernel():
    kernel = xfilter.generate_orthogonal_kernel((3, 5), "bartlett")
    assert kernel.shape == (3, 5)
    np.testing.assert_allclose(kernel[1], [0.0, 0.5, 1.0, 0.5, 0.0])

    kernel = xfilter.generate_orthogonal_kernel([2.8], "ones", fill_value=-1)
    np.testing.assert_allclose(kernel, [-1, 1, 1, -1])


@pytest.mark.parametrize(
    "kernel, data, oshape, odims, osum",
    [
        (3, da1d, (3,), ('nt',), 3),
        ((4, 3), da2d, (4, 3), ('ny', 'nx'), 12),
        ([1, 2, 1], da1d, (3,), ('nt',), 4),
        ([1, 2, 1], da2d, (3, 3), ('ny', 'nx'), 16),
        ({'nx': [1, 2, 1], 'ny': [1, 1, 2, 1, 1]}, da2d, (5, 3), ('ny', 'nx'), 24),
        (np.ones((3, 5)), da2d, (3, 5), ('ny', 'nx'), 15),
        (xr.DataArray(np.ones((3, 5)), dims=('nx', 'ny')), da2d, (5, 3), ('ny', 'nx'), 15),
        (xr.DataArray(np.ones((3, 5)), dims=('ny', 'nx')), da3d, (1, 3, 5), ('nt', 'ny', 'nx'), 15),
        (xr.DataArray(np.ones(3), dims='nx'), da2d, (3,), ('nx',), 3),
    ],
)
def test_filter_generate_kernel(kernel, data, oshape, odims, osum):
    okernel = xfilter.generate_kernel(kernel, data)
    assert okernel.shape == oshape
    assert okernel.dims == odims
    np.testing.assert_allclose(okernel.data.sum(), osum)


def test_filter_convolve():

    da = xr.DataArray(np.ones((5, 5)), dims=('ny', 'nx'))
    da[2, 2] = np.nan

    dac = xfilter.convolve(da, 3, normalize=True, na_thres=1)
    np.testing.assert_allclose(dac.data, np.ones(da.shape))

    dac = xfilter.convolve(da, 3, normalize=True, na_thres=0)
    assert dac.isnull().all()

    dac = xfilter.convolve(da, 3, normalize=False, na_thres=1)
    assert dac.sum() == 160


def test_filter_erode_mask():

    da_in = xr.DataArray(np.random.normal(size=(2, 9, 11)), dims=('nt', 'ny', 'nx'))
    nt, ny, nx = da_in.shape
    i0, n = 2, 5
    da_in[:, i0 : i0 + n, i0 : i0 + n] = np.nan

    # iteration
    da_out_iter = xfilter.erode_mask(da_in, 2)
    assert int(da_out_iter[1].count()) == (da_in[1].size - (n - 4) ** 2)

    # until mask
    mask = np.isnan(da_out_iter)
    da_out_until = xfilter.erode_mask(da_in, mask)
    np.testing.assert_allclose(da_out_iter, da_out_until)


def test_filter_erode_coast():

    da_in = xr.DataArray(np.random.normal(size=(3, 3, 3)), dims=('nt', 'ny', 'nx'))
    da_in[:, 1, 1] = np.nan

    np_out = xfilter.erode_coast(da_in, 1).values
    np_in = da_in.values
    np.testing.assert_allclose(
        np_out[1, 1, 1], 0.25 * (np_in[1, 0, 1] + np_in[1, 2, 1] + np_in[1, 1, 0] + np_in[1, 1, 2])
    )


def test_filter_dermerliac():

    times = np.arange("2000-01-01", "2000-01-10", dtype="M8[h]")
    nt = len(times)
    data = np.cos(np.arange(nt) * np.pi / 12.2)
    data = np.resize(data, (3, 4, nt)).T
    da_in = xr.DataArray(data, dims=('time', 'ny', 'nx'), coords={"time": times})

    da_out = xfilter.demerliac(da_in, na_thres=0)
    assert da_out.dims == da_in.dims
    assert da_out.shape == da_in.shape
    assert da_out.min() > -1e-4
    assert da_out.max() < 1e-4

    da_out = xfilter.demerliac(da_in, na_thres=1)
    assert abs(da_out.mean()) < 0.1


@pytest.mark.parametrize(
    "radius, expected",
    [
        (0.0, [True, True, True, True, True]),
        (0.025, [True, False, True, True, False]),
        (np.inf, [True, False, False, False, False]),
    ],
)
def test_filter_get_decimate_arg(radius, expected):

    x = np.array([0.0, 1.0, 2.0, 5.0, 6.0])
    y = x.copy()
    keep = xfilter._get_decimate_arg_(x, y, radius)
    np.testing.assert_allclose(keep, expected)


def test_filter_decimate():

    ds = xr.Dataset(
        {"temp": ("npts", 20 + np.arange(5))},
        coords={
            "lon": ("npts", [0.0, 1.0, 2.0, 5.0, 6.0]),
            "lat": ("npts", [0.0, 1.0, 2.0, 5.0, 6.0]),
        },
    )

    dsc0 = xfilter.decimate(ds, 160e3, method="pick")
    assert dsc0.lon.dims == ("npts",)
    np.testing.assert_allclose(dsc0.lon.values, [0.0, 2.0, 5.0])

    dac = xfilter.decimate(ds.temp, 160e3, method="pick")
    np.testing.assert_allclose(dac.lon.values, [0.0, 2.0, 5.0])

    dsc1 = xfilter.decimate(ds.drop_vars("temp"), 160e3, method="pick")
    np.testing.assert_allclose(dsc1.lon.values, [0.0, 2.0, 5.0])

    dsc0a = xfilter.decimate(ds, 160e3, method="average", smooth_factor=0)
    xr.testing.assert_equal(dsc0.temp, dsc0a.temp)
    dsc0b = xfilter.decimate(ds, 160e3, method="average", smooth_factor=np.inf)
    assert np.isclose(ds.temp.values.mean(), dsc0b.temp.values.mean())



# -*- coding: utf8 -*-
"""
Kriging adapted from vacumm's module
https://github.com/VACUMM/vacumm/blob/master/lib/python/vacumm/misc/grid/kriging.py

"""
from __future__ import absolute_import
from multiprocessing import Pool, cpu_count
import warnings

import numpy as np
import xarray as xr

from .__init__ import XoaError, xoa_warn
from . import misc
from . import geo as xgeo
from . import coords as xcoords


def _get_blas_func_(name):
    import scipy.linalg.blas

    return scipy.linalg.blas.get_blas_funcs(name)


def _get_lapack_func_(name):
    import scipy.linalg.lapack

    return scipy.linalg.lapack.get_lapack_funcs(name)


def _dgemv_(a, x):
    blas_dgemv = _get_blas_func_('gemv')
    return blas_dgemv(1.0, a, x)


def _symm_(a, b):
    blas_dgemm = _get_blas_func_('gemm')
    return blas_dgemm(1.0, a, b)


def _syminv_(a):
    """Invert a real symmetric definite matrix"""
    return np.linalg.pinv(a)
    n = a.shape[0]
    jj, ii = np.triu_indices(n)
    up = a[jj, ii]
    pptri = _get_lapack_func_("pptri")
    res = pptri(n, up)
    if isinstance(res, tuple):
        info = res[1]
        if info:
            raise KrigingError(f'Error during call to Lapack DPPTRI (info={info})')
        return res[0]
    else:
        return res


class KrigingError(XoaError):
    pass


def get_xyz(obj):
    """Get lon/lat coordinates and data values from a data array or dataset

    Parameters
    ----------
    obj: xarray.DataArray, xarray.Dataset
        If a data array, it must have valid longitude and latitude coordinates.
        If a dataset, it must have a single variable as in the data array case.

    Return
    ------
    numpy.array
        Longitudes as 1D array
    numpy.array
        Latitudes as 1D array
    numpy.array
        Values as a 1D or 2D. None if `obj` is a dataset.
    """
    # Xarray stuff
    obj = xcoords.geo_stack(obj, "npts")
    lon = xcoords.get_lon(obj)
    lat = xcoords.get_lat(obj)

    # Numpy
    x = obj.coords[lon.name].values
    y = obj.coords[lat.name].values
    if isinstance(obj, xr.DataArray):
        z = obj.values.reshape(-1, x.size)
    else:
        z = None

    return x, y, z


def empirical_variogram(
    da: xr.DataArray,
    nbin=30,
    nbin0=10,
    nmax=1500,
    dist_units="m",
    distmax=None,
    errfunc=None,
):
    """Compute the semi-variogram from data

    Parameters
    ----------

    da: xarray.dataArray
        Data array with lon and lat coordinates.
    nmax: optional
        Above this number, size of the sample is reduced by a crude undersampling.
    binned: optional
        If set to a number,
        data are arranged in bins to estimate
        variogram. If set to ``None``, data are
        arranged in bins if the number of pairs
        of points is greater than ``nbindef*nbmin``.
    nbindef: optional
        Default number of bins (not used if ``binned`` is a number).
    nbin0: optional
        If set to a number > 1,
        the first bin is split into nbin0 sub-bins.
        If set to ``None``, it is evaluated with
        ``min(bins[1]/nbmin, nbin)``.
    nbmin: optional
        Minimal number of points in a bin.
    dist_units: str, int, xoa.geo.distance_units
        Distance units as one of: {xgeo.distance_units.rst_with_links}
    distmax: optional
        Max distance to consider.
    errfunc: optional
        Callable function to compute "errors" like square
        root difference between to z values. It take two arguments and
        defaults to :math:`(z1-z0)^2/2`.

    Return
    ------
    xarray.DataArray
        Values as 1D array with name "semivariogram" and with "dist" as distance coordinate in km

    """
    da = xcoords.geo_stack(da, "npts", rename=True, reset_index=True)
    npts = da.sizes["npts"]

    # Undepsample?
    if npts > nmax:
        samp = npts / nmax
        da = da.isel(npts=slice(None, None, samp))
        npts = da.sizes["npts"]

    # Distances
    dist_units = xgeo.distance_units[dist_units]
    dd = xgeo.get_distances(da).values
    if dist_units == xgeo.distance_units.km:
        dd *= 1e-3
    iitriu = np.triu_indices(dd.shape[0], 1)
    d = dd[iitriu]
    del dd

    # Max distance
    if distmax:
        iiclose = d <= distmax
        d = d[iiclose]
        # v = v[valid]
        # del valid
    else:
        iiclose = ...

    # Variogram
    if errfunc is None:

        def errfunc(a0, a1):
            return 0.5 * (a1 - a0) ** 2

    z = np.atleast_2d(da.values)
    v = np.asarray([errfunc(*np.meshgrid(z[i], z[i]))[iitriu][iiclose] for i in range(z.shape[0])])

    # Unique
    d, iiuni = np.unique(d, return_index=True)
    v = v[:, iiuni]

    # Compute edges
    # - classic bins
    nbin = min(d.size, nbin)
    iiedges = np.linspace(0, d.size - 1, nbin + 1).astype('l').tolist()
    # - more details in the first bin
    if nbin0 > 1 and nbin0 < iiedges[1]:  # split first bin
        iiedges = np.linspace(0.0, iiedges[1], nbin0 + 1).astype('l')[:-1].tolist() + iiedges[1:]
        nbin = nbin - 1 + nbin0  # len(iiedges)-1

    # Compute histogram
    db = np.empty(nbin)
    vb = np.empty(nbin)
    for ib in range(nbin):
        iib = slice(iiedges[ib], iiedges[ib + 1] + 1)
        db[ib] = d[iib].mean()
        vb[ib] = v[:, iib].mean()

    # Dataarray
    # dist = xr.DataArray(db, dims="dist", attrs={'long_name': "Distance", "units": "km"})
    attrs = {}
    if "long_name" in da.attrs:
        attrs = {"long_name": "Semi-variogram of " + da.attrs["long_name"]}
    else:
        attrs = {"long_name": "Semi-variogram"}
    if "units" in da.attrs:  # TODO: pint support
        attrs = {"units": da.attrs["units"] + "^2"}
    return xr.DataArray(
        vb,
        dims="dist",
        coords={"dist": ("dist", db, {"long_name": "Distance", "units": str(dist_units)})},
        attrs=attrs,
        name="semivariogram",
    )


empirical_variogram.__doc__ = empirical_variogram.__doc__.format(**locals())


class variogram_model_types(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Supported types of variograms"""

    #: Exponential (default)
    exponential = 1
    #: Linear
    linear = 0
    #: Gausian
    gaussian = 2
    #: Spherical
    spherical = 3


def get_variogram_model_func(mtype, n, s, r, nrelmax=0.2):
    """Get the variogram model function from its name"""

    mtype = variogram_model_types[mtype]

    n = max(n, 0)
    n = min(n, nrelmax * s)
    r = max(r, 0)
    s = max(s, 0)

    if mtype.name == 'linear':
        return lambda h: n + (s - n) * ((h / r) * (h <= r) + 1 * (h > r))

    if mtype.name == 'exponential':
        return lambda h: n + (s - n) * (1 - np.exp(-3 * h / r))

    if mtype.name == 'gaussian':
        return lambda h: n + (s - n) * (1 - np.exp(-3 * h ** 2 / r ** 2))

    if mtype.name == 'spherical':
        return lambda h: n + (s - n) * ((1.5 * h / r - 0.5 * (h / r) ** 3) * (h <= r) + 1 * (h > r))


class VariogramModel(object):
    """Class used when fitting a variogram model to data to better control params

    Parameters
    ----------
    mtype: int, str, variogram_model_types
    dist_units: int, str, xoa.geo.distance_units
    **frozen_params:
        Varigram paramaters that must be frozen.

    """

    param_names = list(get_variogram_model_func.__code__.co_varnames[1:])
    param_names.remove('nrelmax')

    def __init__(self, mtype, dist_units="m", **frozen_params):
        self._dist_units = xgeo.distance_units[dist_units]
        self.mtype = variogram_model_types[mtype]
        self._frozen_params = {}
        self._estimated_params = {}
        self._fit = None
        self._fit_err = None
        self.set_params(**frozen_params)
        self._ev = None

    def __str__(self):
        clsname = self.__class__.__name__
        mtype = self.mtype.name
        dist_units = self._dist_units
        sp = []
        for name in self.param_names:
            sp.append("{}={}".format(name, self[name]))
        sp = ', '.join(sp)
        return f"<{clsname}('{mtype}', dist_units='{dist_units}', {sp})>"

    def __repr__(self):
        return str(self)

    @property
    def dist_units(self):
        """Distance units of type :class:`~xoa.geo.distance_units`"""
        return self._dist_units

    @property
    def frozen_params(self):
        """Frozen parameters"""
        return dict(
            (name, self._frozen_params[name])
            for name in self.param_names
            if name in self._frozen_params
        )

    def get_estimated_params(self):
        """Get parameters that were estimated (not frozen)"""
        return dict(
            (name, self._estimated_params.get(name))
            for name in self.param_names
            if name not in self._frozen_params
        )

    def set_estimated_params(self, overwrite=True, **params):
        """Set the value to no frozen paramters"""
        params_update = dict(
            (name, params[name])
            for name in self.param_names
            if name not in self._frozen_params
            and name in params
            and (not overwrite or name not in self._estimated_params)
        )
        self._estimated_params.update(params_update)

    estimated_params = property(
        get_estimated_params, set_estimated_params, doc='Estimated paramaters as :class:`dict`'
    )

    def set_params(self, **params):
        """Freeze some parameters"""
        params = dict(
            [(p, v) for (p, v) in params.items() if p in self.param_names and v is not None]
        )
        self._frozen_params.update(params)

    def get_params(self, **params):
        """Get current parameters with optional update

        Parameters
        ----------
        params:
            Extra parameters to alter currents values

        Return
        ------
        dict, numpy.array
        """
        these_params = dict(**self.frozen_params, **self.estimated_params)
        if params:
            these_params.update(
                dict(
                    [(p, v) for (p, v) in params.items() if p in self.param_names and v is not None]
                )
            )
        return dict((name, these_params[name]) for name in self.param_names)

    def get_param(self, name):
        """Get a single parameter

        Parameters
        ----------
        name: str
            A valid parameter name

        Return
        ------
        float, None
            Returns None if the parameter is not frozen and has not been estimated yet.
        """
        if name not in self.param_names:
            raise KrigingError(
                f"Invalid param name: {name}. Please use one of: " + ", ".join(self.param_names)
            )
        if name in self._frozen_params:
            return self._frozen_params[name]
        return self._estimated_params.get(name)

    __getitem__ = get_param

    def get_params_array(self):
        """Get the :attr:`estimated_params` as an array

        Return
        ------
        numpy.array
        """
        pp = list(self.estimated_params.values())
        if None in pp:
            raise VariogramModelError(
                "Not all parameters are estimated: {}".format(self.estimated_params)
            )
        return np.array(pp)

    def set_params_array(self, pp):
        """Set the :attr:`estimated_param` with an array

        Parameters
        ----------
        pp: numpy.array
            Array of estimated parameters
        """
        for i, name in enumerate(self.estimated_params):
            self._estimated_params[name] = pp[i]
        return self.params

    @property
    def params(self):
        """Current variogram model parameters"""
        return self.get_params()

    def apply(self, d, pp=None):
        """Call the variogram model function

        Parameters
        ----------
        d: array
            Distances
        """
        return self.get_func(pp)(d)

    __call__ = apply

    def get_func(self, pp=None):
        """Get the variogram model function using `pp` variable arguments"""
        if pp is not None:
            params = self.set_params_array(pp)
        else:
            params = self.get_params()
            if None in list(params.values()):
                raise VariogramModelError(
                    "Not all parameters are estimated: {}".format(self.estimated_params)
                )
        return get_variogram_model_func(self.mtype, **params)

    def fit(self, da: xr.DataArray, **kwargs):
        """Estimate parameters from data"""
        # We need a data array
        if isinstance(da, xr.Dataset):
            if len(da) == 1:
                da = da[list(da)[0]]
            elif len(da) > 1:
                xoa_warn(
                    "Multiple candidate variables found in the dataset for estimating "
                    "the variogram parameters. Keeping only the first one."
                )
            elif len(da) == 0:
                raise KrigingError(
                    "No variable found in the dataset for estimating the variogram parameters."
                )

        # Empirical variogram
        if da.name == "semivariogram" or da.name == "variogram" and "dist" in da.coords:
            if (
                "units" in da.dist.attrs
                and da.dist.attrs["units"] in xgeo.distance_units
                and xgeo.distance_units[da.dist.attrs["units"]] != self._dist_units
            ):
                xoa_warn(
                    "Incompatible distance units: {} vs {}".format(
                        self._dist_units, da.dist.attrs["units"]
                    )
                )
            ev = da
        else:
            kwargs["dist_units"] = self._dist_units
            ev = empirical_variogram(da, **kwargs)
        dist = ev.dist.values
        values = ev.values
        self._ev = ev

        # First guess of paramaters
        imax = np.ma.argmax(values)
        self.set_estimated_params(n=0.0, s=values[imax], r=dist[imax], overwrite=False)
        pp0 = self.get_params_array()

        # Fitting
        from scipy.optimize import minimize

        def func(pp):
            return ((values - self(dist, pp)) ** 2).sum()

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'divide by zero encountered in divide')
            self._fit = minimize(
                func, pp0, bounds=[(np.finfo('d').eps, None)] * len(pp0), method='L-BFGS-B'
            )
            pp = self._fit['x']
            self._fit_err = np.sqrt(func(pp)) / values.size

        self.set_params_array(pp)

    def plot(self, rmax=None, nr=100, show_params=True, **kwargs):
        """Plot the semivariogram

        Parameters
        ----------
        rmax: float
            Max range in meters
        nr: int
            Number of points to plot the curve
        show_params: bool, dict
            Show a text box that contains the variogram parameters in the lower right corner.
        kwargs: dict
            Extra keyword are passed to the `xarray.DataArray.plot` callable accessor
        """
        # Distances
        if rmax is None and self._ev is not None:
            rmax = self._ev.dist.max()
        else:
            rmax = self["r"] * 1.2
        dist = np.linspace(0, rmax, nr)

        # Array and plot
        du = str(self._dist_units)
        mv = xr.DataArray(
            self.get_func()(dist),
            dims='dist',
            coords=[("dist", dist, {"long_name": "Distance", "units": du})],
            attrs={"long_name": self.mtype.name.title() + " fit"},
        )
        kwargs.setdefault("label", mv.long_name)
        p = mv.plot(**kwargs)

        # Text box for params
        if show_params:
            params = self.params.copy()
            text = [
                "r[ange] = {:<g} {}".format(params["r"], du),
                "n[ugget] = {:<g}".format(params["n"]),
                "s[ill] = {:<g}".format(params["s"]),
            ]
            maxlen = max([len(t) for t in text])
            text = "\n".join(t.ljust(maxlen) for t in text)
            axes = p[0].axes
            axes.text(
                0.98,
                0.04,
                text,
                transform=axes.transAxes,
                family="monospace",
                bbox=dict(facecolor=(1, 1, 1, 0.5)),
                ha="right",
            )
        return p


class VariogramModelError(KrigingError):
    pass


class kriging_types(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Supported kriging of variograms"""

    #: Exponential (default)
    ordinary = 1
    #: Linear
    simple = 0


class Kriger(object):
    """Kriger that supports clusterization to limit memory

    Big input cloud of points (size > ``npmax``)
    are split into smaller clusters using cluster analysis of distance with
    function :func:`~xoa.geo.clusterize`.

    The problem is solved in this way:

        #. Input points are split in clusters if necessary.
        #. The input variogram matrix is inverted
           for each cluster, possibly using
           :mod:`multiprocessing` if ``nproc>1``.
        #. Values are computed at output positions
           each using the inverted matrix of cluster.
        #. Final value is a weighted average of
           the values estimated using each cluster.
           Weights are inversely proportional to the inverse
           of the squared error.

    Parameters
    ----------

    da: xarray.DataArray, xarray.Dataset
        Input positions and optionaly data.
    krigtype: optional
        Kriging type: {kriging_types.rst_with_links}.
    variogram_func: callable, VariogramModel, optional
        Callable to be used as a variogram function.
        It is either a function or an instance of :class:`VariogramModel`.
    npmax: optional
        Maximal number of points to be used simultanously for kriging.
        When the number of input points is greater than this value,
        clusterization is applied.
    nproc: optional
        Number of processes to use to invert matrices.
        Set it to a number <2 to switch off parallelisation.
    exact: optional
        If True, variogram is exactly zero when distance is zero.


    """

    def __init__(
        self,
        da: xr.DataArray,
        krigtype,
        variogram_func,
        npmax=None,
        nproc=None,
        exact=False,
        dist_units=None,
        mean=None,
        farvalue=None,
        **kwargs,
    ):
        # Kriging type
        self.krigtype = kriging_types[krigtype]

        # Variogram function
        if isinstance(variogram_func, str):
            variogram_func = VariogramModel(variogram_func, dist_units=dist_units)
        if isinstance(variogram_func, VariogramModel):
            if dist_units is None:
                dist_units = variogram_func.dist_units
            else:
                dist_units = xgeo.distance_units[dist_units]
            if dist_units != variogram_func.dist_units:
                xoa_warn(
                    f"Incompatible distance units: {dist_units} vs {variogram_func.dist_units}"
                )
            variogram_func.fit(da)
            self._variogram_func = variogram_func
        else:
            self._variogram_func = variogram_func
        self._dist_units = xgeo.distance_units[dist_units]

        # Clusters
        if npmax is None:
            npmax = np.inf
        self._clusters = xgeo.clusterize(da, npmax=npmax, split=True)
        self._unstacked_coords = {}
        for cname, cdat in self.clusters[0].coords.items():
            if "npts" not in cdat.dims:
                self._unstacked_coords[cname] = cdat

        # Number of cores for parallel computing
        if nproc is None:
            nproc = cpu_count()
        else:
            nproc = max(1, min(cpu_count(), nproc))
        self.nproc = min(nproc, self.nclust)

        # Other parameters
        self.exact = exact
        if self.krigtype != kriging_types.simple:
            mean = 0.0
        elif mean is None:
            mean = float(da.mean())
        self.mean = mean

    @property
    def clusters(self):
        return self._clusters

    @property
    def nclust(self):
        """Number of clusters"""
        return len(self.clusters)

    @property
    def npmax(self):
        """Max number of points per cluster"""
        return self.clusters[0].npmax

    @property
    def dist_units(self):
        """Distance units"""
        return self._dist_units

    @property
    def variogram_func(self):
        """Variogram function or callable, like a :class:`VariogramModel` instance"""
        return self._variogram_func

    @property
    def Ainv(self):
        """Get the inverse of the A matrix"""

        # Already computed
        if hasattr(self, '_Ainv'):
            return self._Ainv

        # Variogram function
        vgf = self.variogram_func

        # Loop on clusters
        Ainv = []
        AA = []
        plus1 = int(self.krigtype == kriging_types.ordinary)
        for ic, clust in enumerate(self.clusters):

            # Get distance between input points
            dd = xgeo.get_distances(clust).values

            # Form A
            npts = clust.sizes["npts"]
            A = np.empty((npts + plus1, npts + plus1))
            A[:npts, :npts] = vgf(dd)
            if self.exact:
                np.fill_diagonal(A, 0)
                A[:npts, :npts][np.isclose(A[:npts, :npts], 0.0)] = 0.0
            if self.krigtype == kriging_types.ordinary:
                A[-1] = 1
                A[:, -1] = 1
                A[-1, -1] = 0

            # Invert for single processor
            if self.nproc == 1:
                Ainv.append(_syminv_(A))
            else:
                AA.append(A)

        # Multiprocessing inversion
        if self.nproc > 1:
            pool = Pool(self.nproc)
            Ainv = pool.map(_syminv_, AA, chunksize=1)
            pool.close()

        # Fortran arrays
        Ainv = [np.asfortranarray(ainv, 'd') for ainv in Ainv]
        self._Ainv = Ainv
        return Ainv

    def interp(self, dso: xr.Dataset, block=None, name=None):
        """Interpolate to the `dso` positions

        Parameters
        ----------
        dso: xr.Dataset
            Dataset that contains lon and lat coordinates
        block: None, int
            Number of nearest neighbours for block kriging
        name: str, None:
            Default name of the output data array.

        Return
        ------
        xarray.Dataset
            A dataset that contains the interpolated values and associated erreors

        """

        # Inits
        dso = xcoords.geo_stack(dso, rename=False, drop=True)
        sname, xname, yname = dso.encoding["geo_stack"]
        stacked_coords = {}
        for cname, cdat in dso.coords.items():
            if sname in cdat.dims:
                stacked_coords[cname] = cdat
        vgf = self.variogram_func
        nptso = dso.sizes[sname]
        vname = self.clusters[0].encoding["clust_var_names"][0]
        so = self.clusters[0][vname].shape[:-1] + (nptso,)
        dimso = self.clusters[0][vname].dims[:-1] + (sname,)
        dimsoe = (sname,)
        zo = np.zeros(so, 'd')
        eo = np.zeros(nptso, 'd')
        wo = np.zeros(nptso, 'd')
        if block:
            xyo = np.dstack([dso[xname].values, dso[yname].values])[0]

        # Loop on clusters
        Ainv = self.Ainv
        plus1 = int(self.krigtype == kriging_types.ordinary)
        for ic in range(self.nclust):  # TODO: multiproc here?

            # Distances to output points
            dd = xgeo.get_distances(self.clusters[ic], dso, units=self._dist_units).values

            # Form B
            npts = self.clusters[ic].sizes["npts"]
            B = np.empty((npts + plus1, nptso))
            B[:npts] = vgf(dd)
            if self.krigtype == kriging_types.ordinary:
                B[-1] = 1
            if self.exact:
                B[:npts][np.isclose(B[:npts], 0.0)] = 0.0
            del dd

            # Block kriging
            if block:
                from scipy.spatial import cKDTree

                tree = cKDTree(xyo)
                Bb = B.copy()
                for i, iineigh in enumerate(tree.query_ball_tree(tree, block)):
                    Bb[:, i] = B[:, iineigh].mean()
                B = Bb

            # Compute weights
            W = np.ascontiguousarray(_symm_(Ainv[ic], np.asfortranarray(B, 'd')))

            # Interpolate
            zc = self.clusters[ic][vname].values
            if self.krigtype == kriging_types.simple:
                zc = zc - self.mean
            z = zc @ W[:npts]
            if self.krigtype == kriging_types.simple:
                z += self.mean

            # Get error
            #            e = (W[:-1]*B[:-1]).sum(axis=0)
            e = (W * B).sum(axis=0)
            del W, B

            # Weigthed contribution based on errors
            w = 1 / e ** 2
            if self.nclust > 1:
                z[:] *= w
            wo += w
            del w
            zo += z
            del z

        # Error
        eo = 1 / np.sqrt(wo)

        # Normalization
        if self.nclust > 1:
            zo /= wo

        # Format
        coords = self._unstacked_coords.copy()
        coords.update(stacked_coords)
        dao = xr.DataArray(zo, dims=dimso, coords=coords, attrs=self.clusters[0].attrs)
        daoe = xr.DataArray(
            eo, dims=dimsoe, coords=stacked_coords, attrs={"long_name": "Squared error"}
        )
        if name is None:
            name = self.clusters[0].encoding["clust_var_names"][0]
        if name is None:
            name = "data"
        return xr.Dataset({name: dao, name + "_error": daoe}).unstack()

        # gc.collect()
        # if geterr:
        #     return zo, eo
        # return zo

    __call__ = interp


Kriger.__doc__ = Kriger.__doc__.format(**locals())


def krig(dsi, dso, krigtype="ordinary", **kwargs):
    """Quickly krig data"""
    kwinterp = {}
    for key in "block", "name":
        if key in kwargs:
            kwinterp[key] = kwargs.pop(key)
    return Kriger(dsi, krigtype, **kwargs).interp(dso, **kwinterp)



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.thermdyn` module
"""

import numpy as np
import xarray as xr

from xoa import thermdyn


def test_thermdyn_mixed_layer_depth():

    # 1d arrays
    depth = xr.DataArray(
        np.linspace(-50, 0.0, 6), dims="z", attrs={"long_name": "Depth", "positive": "up"}
    )
    temp = xr.DataArray(np.linspace(10, 20.0, 6), dims="z", coords={"depth": depth}, name="temp")
    kz = xr.DataArray(np.linspace(0.0005 * 10, 0, 6), dims="z", coords={"depth": depth})

    # temp 1d
    mld = thermdyn.mixed_layer_depth(temp, method="deltatemp")
    np.testing.assert_allclose(mld, 1)
    assert mld.long_name == "Mixed layer depth"
    assert mld.dims == ()

    # implicit
    mld = thermdyn.mixed_layer_depth(temp)
    np.testing.assert_allclose(mld, 1)

    # kz
    mld = thermdyn.mixed_layer_depth(kz, method="kzmax")
    np.testing.assert_allclose(mld, 5)

    # 2d arrays
    temp2d = temp.expand_dims("lon").transpose(..., "lon")
    depth2d = depth.expand_dims("lon").transpose(..., "lon")
    temp2d.coords["depth"] = depth2d
    mld2d = thermdyn.mixed_layer_depth(temp2d, method="deltatemp")
    assert mld2d.dims == ("lon",)
    assert mld2d.shape == (1,)
    np.testing.assert_allclose(mld2d, 1)
